/*
 * DirectShow audio capture for ffmpeg system
 * Copyright (c) 2005 DivX Inc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */

#define AV_DONT_OVERRIDE_STDC
#include "FilterUtil.h"

#include <streams.h>
#include "AudioSinkCallBack.h"

#include "AudioCapture.h"

DShowAudioCaptureGraph::DShowAudioCaptureGraph( AVFormatContext *avctx, const char *device )
	:DShowCaptureGraph(avctx,device)
	,m_device_input(NULL)
{
	char *separ = strrchr( m_devicename, '%');
	if (separ != NULL) {
		m_device_input = av_strdup( separ + 1);
		*separ = '\0';
	}
	else
		m_device_input = NULL;
}


DShowAudioCaptureGraph::~DShowAudioCaptureGraph()
{
	if ( m_device_input != NULL )
		av_free( m_device_input );
	DelFromRot( dwRotId );
}

bool DShowAudioCaptureGraph::CreateGraph( AVFormatParameters & avctx )
{
	// build the graph for the numbered device
    HRESULT hr;
    IBaseFilter *pSrcFilter=NULL;

    // Get DirectShow interfaces
    hr = GetInterfaces();
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Failed to get interfaces!  hr=0x%x\n"), hr);
		return false;
    }

    // Attach the filter graph to the capture graph
    hr = AttachFiltergraph();
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Failed to set capture filter graph!  hr=0x%x\n"), hr);
        return false;
    }

    // Use the system device enumerator and class enumerator to find
    // an audio capture/preview device, such as a desktop USB audio camera.
	hr = FindCaptureDevice(&pSrcFilter);
    if (FAILED(hr))
    {
        // Don't display a message because FindCaptureDevice will handle it
		return false;
    }
   
    // Add Capture filter to our graph.
	hr = g_pGraph->AddFilter(pSrcFilter, L"Audio Capture");
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't add the capture filter to the graph!  hr=0x%x\r\n\r\n") 
            TEXT("If you have a working audio capture device, please make sure\r\n")
            TEXT("that it is connected and is not being used by another application.\r\n\r\n")
            TEXT("The sample will now close."), hr);
        pSrcFilter->Release();
        return false;
    }

	// add a AudioSink
	assert(	m_Sink == NULL );
	// FIXME: get a codec context too !
	m_Sink = new AudioSinkCallBack(*(AVCodecContext*)NULL, avctx.sample_rate);
	// query interface for the sink filter and add the
	// filter to the filtergraph
	IBaseFilter *pSinkFilter;
	if (FAILED(m_Sink->QueryInterface(IID_IBaseFilter, (void **) &pSinkFilter)))
		return false;
	if (FAILED(g_pGraph->AddFilter(pSinkFilter, L"Sink Filter")))
		return false;

    // Render the preview pin on the audio capture filter to our sink
	g_pCapture->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Audio, pSrcFilter, NULL, pSinkFilter);
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't render the audio capture stream.  hr=0x%x\r\n")
            TEXT("The capture device may already be in use by another application.\r\n\r\n")
            TEXT("The sample will now close."), hr);
        pSrcFilter->Release();
        return false;
    }

    // Now that the filter has been added to the graph and we have
    // rendered its stream, we can release this reference to the filter.
    pSrcFilter->Release();

	AddToRot( g_pGraph, dwRotId );
//while(1) Sleep(100);

	return true;
}

size_t DShowAudioCaptureGraph::GetDeviceList( StringList & out_list, AVFormatContext *avctx )
{
	return DShowCaptureGraph::GetDeviceList(out_list, CLSID_AudioInputDeviceCategory, avctx);
}

void DShowAudioCaptureGraph::GetDeviceCaps( const std::wstring & device, 
										   std::vector<WAVEFORMATEX> & out_list, 
										   StringList &input_list, 
										   AVFormatContext *avctx )
{
	out_list.clear();	

	HRESULT hr;
	IBaseFilter *pSrc = FindDevice( device.c_str(), CLSID_AudioInputDeviceCategory, avctx );

	if (pSrc != NULL)
	{
		IEnumPins *pEnumPins;
		IPin *pOutputPin;

		hr = pSrc->EnumPins( &pEnumPins );
		if (FAILED(hr))
		{
			av_log(avctx, AV_LOG_ERROR, TEXT("GetDeviceCaps failed: no pin enumeration !\n"));
			pSrc->Release();
			goto done;
		}

		while( S_OK == pEnumPins->Next( 1, &pOutputPin, NULL ) )
		{
			PIN_INFO info;

			// TODO: add support for IID_IKsPropertySet to check for PIN_CATEGORY_CAPTURE 
			if( S_OK == pOutputPin->QueryPinInfo( &info ) )
			{
				if( info.pFilter ) info.pFilter->Release();
				if( info.dir == PINDIR_INPUT )
				{
					input_list.push_back( info.achName );
					pOutputPin->Release();
					continue;
				}
				av_log(avctx, AV_LOG_DEBUG, "GetDeviceCaps: trying pin %S\n", info.achName );
			}

			IEnumMediaTypes *p_enum_mt;
			if ( SUCCEEDED(pOutputPin->EnumMediaTypes(&p_enum_mt)))
			{
				if ( SUCCEEDED(p_enum_mt->Reset()))
				{
					AM_MEDIA_TYPE *p_mt;
					ULONG          fetched;
					while( S_OK == p_enum_mt->Next( 1, &p_mt, &fetched ) )
					{
						if (pOutputPin->QueryAccept( p_mt ) == S_OK )
						{
							if( MEDIATYPE_Audio == p_mt->majortype && FORMAT_WaveFormatEx == p_mt->formattype )
							{
								WAVEFORMATEX *pFex = (WAVEFORMATEX *)p_mt->pbFormat;
								if (pFex->wFormatTag == 1 && pFex->wBitsPerSample == 16) // only support PCM 16 bits
									out_list.push_back(*pFex);
							}
						}
						DeleteMediaType( p_mt );
					}
				}
				p_enum_mt->Release();
			}
			pOutputPin->Release();

		}

		pEnumPins->Release();
		pSrc->Release();
	}

done:
	;
}

HRESULT DShowAudioCaptureGraph::FindCaptureDevice(IBaseFilter ** ppSrcFilter)
{
    HRESULT hr = S_FALSE;
	IBaseFilter * pSrc = NULL;

    if (!ppSrcFilter)
        return E_POINTER;
   
	wchar_t p_buf[256];
	int i_convert = MultiByteToWideChar(CP_ACP, 0, m_devicename, strlen(m_devicename)+1, p_buf, sizeof(p_buf));

	pSrc = FindDevice( p_buf, CLSID_AudioInputDeviceCategory, m_avctx );

    if (pSrc != NULL )
    {
		// select the right input pin
		IPin *pInputPin;
		IEnumPins *pEnumPins;

		hr = pSrc->EnumPins( &pEnumPins );
		if (FAILED(hr))
		{
			av_log(m_avctx, AV_LOG_ERROR, TEXT("FindCaptureDevice failed: no pin enumeration !\n"));
			return hr;
		}

		while( S_OK == pEnumPins->Next( 1, &pInputPin, NULL ) )
		{
			PIN_INFO info;
			IAMAudioInputMixer *pInputSelect;

			if (SUCCEEDED(pInputPin->QueryInterface(IID_IAMAudioInputMixer , (void **) &pInputSelect)))
			{
				// TODO: add support for IID_IKsPropertySet to check for PIN_CATEGORY_CAPTURE 
				if( S_OK == pInputPin->QueryPinInfo( &info ) )
				{
					char p_buf[256];
					WideCharToMultiByte( CP_ACP, 0, info.achName, wcslen(info.achName)+1, p_buf, sizeof(p_buf), NULL, NULL );

					if( strcmp(p_buf, m_device_input) == 0 ) {
						pInputSelect->put_Enable(true);
						av_log(m_avctx, AV_LOG_DEBUG, "FindCaptureDevice: selecting input pin %S\n", info.achName );
					} else
						pInputSelect->put_Enable(false);
					if( info.pFilter ) info.pFilter->Release();
				}
				pInputSelect->Release();
			}
			pInputPin->Release();
		}

		pEnumPins->Release();
    }
    else
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Unable to access audio capture device!\n"));   
        return E_FAIL;
    }

    // Copy the found filter pointer to the output parameter.
    // Do NOT Release() the reference, since it will still be used
    // by the calling function.
    *ppSrcFilter = pSrc;

    return hr;
}

void DShowAudioCaptureGraph::Configure( const wchar_t *device, HWND hwParent )
{
	IBaseFilter *pSrc = FindDevice( device, CLSID_AudioInputDeviceCategory, NULL );
	if (pSrc == NULL)
		return;
	ISpecifyPropertyPages *pProp;
	HRESULT hr = pSrc->QueryInterface(IID_ISpecifyPropertyPages, (void **)&pProp);
	if (SUCCEEDED(hr)) 
	{
		// Get the filter's name and IUnknown pointer.
		FILTER_INFO FilterInfo;
		hr = pSrc->QueryFilterInfo(&FilterInfo); 
		IUnknown *pFilterUnk;
		pSrc->QueryInterface(IID_IUnknown, (void **)&pFilterUnk);

		// Show the page. 
		CAUUID caGUID;
		pProp->GetPages(&caGUID);
		pProp->Release();
		OleCreatePropertyFrame(
			hwParent,               // Parent window
			0, 0,                   // Reserved
			FilterInfo.achName,     // Caption for the dialog box
			1,                      // Number of objects (just the filter)
			&pFilterUnk,            // Array of object pointers. 
			caGUID.cElems,          // Number of property pages
			caGUID.pElems,          // Array of property page CLSIDs
			0,                      // Locale identifier
			0, NULL                 // Reserved
		);

		// Clean up.
		pFilterUnk->Release();
		CoTaskMemFree(caGUID.pElems);
	}
	pSrc->Release();
}

int DShowAudioCaptureGraph::ReadFrame(AVPacket *pkt)
{
	if ( g_psCurrent == Stopped )
	{
		if (FAILED(g_pMC->Run()))
			return -1;

		g_psCurrent = Running;
	}

	// prepare for 1s of audio
	int buffer_size = m_Stream->codec->bits_per_sample * m_Stream->codec->channels * m_Stream->codec->sample_rate / 8;
    if (av_new_packet(pkt,buffer_size) < 0) {
        return AVERROR_NOMEM;
    }

	m_Sink->SetOutput( pkt->data, buffer_size, &pkt->pts );
	while (buffer_size == 0)
		Sleep(20); // wait for data dirty hack ?

	pkt->size = buffer_size;
	pkt->stream_index = m_Stream->index;

	return buffer_size;
}

bool DShowAudioCaptureGraph::FillCodec( AVFormatParameters & avfctx )
{
	m_Stream = av_new_stream(m_avctx, 0);
	AVCodecContext & avctx = *m_Stream->codec;

	// TODO : make sure the values are correct
	avctx.codec_type = CODEC_TYPE_AUDIO;
	avctx.codec_id = CODEC_ID_PCM_S16LE;
	avctx.sample_rate = avfctx.sample_rate;
	avctx.channels = avfctx.channels;
	avctx.bits_per_sample = 16;
	m_Stream->time_base = AV_TIME_BASE_Q;

	AVRational_me(avctx.time_base, 1, avfctx.sample_rate);

	avctx.bit_rate = avctx.channels*avctx.sample_rate*2;
	return true;
}

extern "C" {

static int dshow_iread_audio_header(struct AVFormatContext *cont, AVFormatParameters *ap)
{
	DShowAudioCaptureGraph **graph = (DShowAudioCaptureGraph **)cont->priv_data;
	
	if (!ap || ap->sample_rate <= 0 || ap->channels <= 0)
        return -1;

	*graph = new DShowAudioCaptureGraph(cont, ap->device);

	if (!(*graph)->CreateGraph( *ap ))
	{
		delete *graph;
		*graph = NULL;
		return -1;
	}
	(*graph)->FillCodec( *ap );

	cont->pb.is_streamed = 1;

	return 0;
}

static int dshow_iread_packet(struct AVFormatContext *cont, AVPacket *pkt)
{
	DShowAudioCaptureGraph **graph = (DShowAudioCaptureGraph **)cont->priv_data;
	return (*graph)->ReadFrame( pkt );
}

static int dshow_iread_close(struct AVFormatContext *cont)
{
	DShowAudioCaptureGraph **graph = (DShowAudioCaptureGraph **)cont->priv_data;
	delete *graph;
	return 0;
}

static AVInputFormat dshow_acap_format = {
    "dshowacap",
    "DirectShow audio capture",
    sizeof(DShowAudioCaptureGraph*),
    NULL, // read_probe
    dshow_iread_audio_header,
    dshow_iread_packet,
    dshow_iread_close,
    NULL, // read_seek
    NULL, // read_timestamp <-
	AVFMT_NOFILE,
	/* .extensions =*/ "acap",
};

int dshow_audio_cap_init(void)
{
    av_register_input_format(&dshow_acap_format);
    return 0;
}

}; // extern "C"
