/*
 * Generic DirectShow decoder (for win32 only)
 * Copyright (c) 2006 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file DShowDecoder.cpp
 * generic DShow entry point.
 */

#include <string>

#include <streams.h>

#include "ffconfig.h"
#include "common.h"
#include "avformat.h"

#if defined(CONFIG_WIN32)

#include "FilterUtil.h"
#include "VideoSinkCallBack.h"
#include "DShowDecoder.h"

DShowDecoder::~DShowDecoder()
{
    if (m_oGraphControl != NULL)
    {
        m_oGraphControl->Stop();

        HRESULT hr;
        OAFilterState filter_state;
        hr = m_oGraphControl->GetState( 1000, &filter_state );

        DelFromRot( dwRotId );
        m_oGraph->RemoveFilter( pSinkFilter );
        m_oGraph->RemoveFilter( pSourceFilter );

        // remove remaining filters
        IEnumFilters *pFilters;
        if (SUCCEEDED(m_oGraph->EnumFilters(&pFilters))) {
            IBaseFilter *pFilter;
            ULONG    n;
            while (pFilters->Next(1, &pFilter, &n) == S_OK) {
                m_oGraph->RemoveFilter( pFilter );
            }
        }

        hr = m_oGraphControl->GetState( 1000, &filter_state );

        m_oGraphControl->Release();
    }

    if (m_oGraph != NULL)
        m_oGraph->Release();

    delete pSource;
    delete pSink;
}

bool DShowDecoder::GivePinStrName(IBaseFilter * pDecoderFilter, LPWSTR& PinOneIdName, LPWSTR& PinTwoIdName)
{
    IEnumPins * pEnumPins;
    if (FAILED(pDecoderFilter->EnumPins(&pEnumPins)))
        return false;
    bool result = false;
    IPin * pPin;
    PIN_INFO pPinInfo;
    if (FAILED(pEnumPins->Next(1, &pPin, NULL)))
        goto done_pin_name;
    pPin->QueryPinInfo(&pPinInfo);
    pPin->QueryId(&PinOneIdName);
    if (FAILED(pEnumPins->Next(1, &pPin, NULL)))
        goto done_pin_name;
    pPin->QueryPinInfo(&pPinInfo);
    pPin->QueryId(&PinTwoIdName);

    result = true;
done_pin_name:
    pEnumPins->Release();
    return result;
}

// connect all the pins of the graph
bool DShowDecoder::ConnectGraph( IGraphBuilder *m_oGraph, IBaseFilter *pSourceFilter, IBaseFilter *pDecoderFilter, IBaseFilter *pSinkFilter )
{
    // Render the preview pin on the audio capture filter
    // Use this instead of g_pGraph->RenderFile
	if (pDecoderFilter == NULL)
	{
		if (FilterUtil::Connect(m_oGraph, pSourceFilter, pSinkFilter, L"Output", L"Input") != S_OK)
			return false;
	}
	else
	{
		if (FilterUtil::Connect(m_oGraph, pSourceFilter, pDecoderFilter, L"Output", NULL) != S_OK)
			return false;
		if (FilterUtil::Connect(m_oGraph, pDecoderFilter, pSinkFilter, NULL, L"Input") != S_OK)
			return false;
	}
	return true;
}

bool DShowDecoder::ConnectGraph( IGraphBuilder *m_oGraph, IBaseFilter *pSourceFilter, IBaseFilter *pSinkFilter )
{
    // Render the preview pin on the audio capture filter
    // Use this instead of g_pGraph->RenderFile
	if (FilterUtil::Connect(m_oGraph, pSourceFilter, pSinkFilter, L"Output", L"Input") != S_OK)
		return false;
	return true;
}

bool DShowDecoder::Deliver( const uint8_t *buf, int buf_size )
{
    return pSource->Deliver( buf, buf_size );
}

bool DShowDecoder::InitGraph( AVCodecContext *avctx, const uint8_t *buf, int buf_size )
{
    bool result = false;
	IBaseFilter *pDecoderFilter = NULL;
	const std::vector<whitelist> & whitelist = GetWhiteList();
	char filter_name[256];
    
    if (FAILED(CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC_SERVER, IID_IGraphBuilder, (void **) &m_oGraph)))
        goto done_init;

    AddToRot( m_oGraph, dwRotId );

	av_log(avctx, AV_LOG_INFO, "Trying to locate a DirectShow decoder\n");

	if (whitelist.size() != 0)
	{
        size_t i;
		av_log(avctx, AV_LOG_INFO, "Using the DirectShow whitelist\n");
		// Get the filter we're going to use to decode
		for (i=0; i<whitelist.size(); i++)
		{
			CharToOemW(whitelist.at(i).name, filter_name);
			av_log(avctx, AV_LOG_INFO, "Trying filter '%s'\n", filter_name);
			if (FilterUtil::LoadFilter(m_oGraph, *(const GUID*)& whitelist.at(i).guid, whitelist.at(i).name, &pDecoderFilter))
			{
				av_log(avctx, AV_LOG_INFO, "Found filter '%s'\n", filter_name);
				break;
			}
		}

		// none of the filters were found from our whitelist
		if (i == whitelist.size()) {
			av_log(avctx, AV_LOG_ERROR, "No filter from the white list were found!\n");
			goto done_init;
		}
	}

    // instantiate my source callback filter
	pSource = GetSource(avctx, buf, buf_size);
	if (pSource == NULL)
        goto done_init;

	av_log(avctx, AV_LOG_DEBUG, "Source filter generated\n");

    // query interface for the source callback filter and add the
    // filter to the filtergraph
    if (FAILED(pSource->QueryInterface(IID_IBaseFilter, (void **) &pSourceFilter)))
        goto done_init;

    if (FAILED(m_oGraph->AddFilter(pSourceFilter, L"Source Filter")))
        goto done_init;

	av_log(avctx, AV_LOG_DEBUG, "Source filter added\n");

    // instantiate my sink callback filter
	pSink = GetSink(avctx, buf, buf_size);
	if (pSink == NULL)
        goto done_init;

	av_log(avctx, AV_LOG_DEBUG, "Sink filter generated\n");

	// query interface for the sink filter and add the
    // filter to the filtergraph
    if (FAILED(pSink->QueryInterface(IID_IBaseFilter, (void **) &pSinkFilter)))
        goto done_init;

    if (FAILED(m_oGraph->AddFilter(pSinkFilter, L"Sink Filter")))
        goto done_init;

	av_log(avctx, AV_LOG_DEBUG, "Sink filter added\n");

    // connect the graph
#if 0 // let the graph builds alone
	if (whitelist.size() != 0)
		result = ConnectGraph( m_oGraph, pSourceFilter, pDecoderFilter, pSinkFilter );
	else
#endif
	result = ConnectGraph( m_oGraph, pSourceFilter, pSinkFilter );

    if (!result)
        goto done_init;

    // run the graph
    if (FAILED(m_oGraph->QueryInterface(IID_IMediaControl, (void **) &m_oGraphControl)))
        goto done_init;

    if (FAILED(m_oGraphControl->Run()))
        goto done_init;

    m_bGraphInit = true;

	GetFilterName(filter_name, sizeof(filter_name));
	av_log(avctx, AV_LOG_INFO, "Using filter '%s'\n", filter_name);

done_init:
    return result;
}

void DShowDecoder::GraphInfo(AVCodecContext *avctx)
{
    ASSERT(m_bGraphInit == true);

    av_log( avctx, AV_LOG_INFO, "DirectShow graph : " );
    // remove remaining filters
    IEnumFilters *pFilters;
    if (SUCCEEDED(m_oGraph->EnumFilters(&pFilters))) {
        IBaseFilter *pFilter;
        FILTER_INFO filter_info;
        ULONG    n;
        int filter_number = 0;
        CHAR pFilterName[1024];

        // count the number of filters
        while (pFilters->Next(1, &pFilter, &n) == S_OK) {
            filter_number++;
            pFilter->Release();
        }
        pFilters->Reset();

        while (pFilters->Next(1, &pFilter, &n) == S_OK) {
            filter_number--;
            if (FAILED(pFilter->QueryFilterInfo( &filter_info )))
                av_log( avctx, AV_LOG_INFO, "unknown" );
            else
            {
                WideCharToMultiByte( CP_UTF8, 0, filter_info.achName, -1, pFilterName, sizeof(pFilterName), NULL, NULL );
                av_log( avctx, AV_LOG_INFO, "%s", pFilterName );
            }
            if (filter_number)
                av_log( avctx, AV_LOG_INFO, " <= " );
            pFilter->Release();
        }

        pFilters->Release();
    }

    av_log( avctx, AV_LOG_INFO, "\n" );
}

void DShowDecoder::GetFilterName(char *string, unsigned int buf_size)
{
    ASSERT(m_bGraphInit == true);

    *string = '\0';
    // remove remaining filters
    IEnumFilters *pFilters;
    if (SUCCEEDED(m_oGraph->EnumFilters(&pFilters))) {
        IBaseFilter *pFilter;
        FILTER_INFO filter_info;
        ULONG    n;
        CHAR pFilterName[1024];

        while (pFilters->Next(1, &pFilter, &n) == S_OK) {
            if (!FAILED(pFilter->QueryFilterInfo( &filter_info )))
            {
                WideCharToMultiByte( CP_UTF8, 0, filter_info.achName, -1, pFilterName, sizeof(pFilterName), NULL, NULL );
                if (strcmp(pFilterName, "Sink Filter") != 0 && strcmp(pFilterName, "Source Filter") != 0)
                {
                    if (string[0] != '\0')
                        pstrcat(string, buf_size, " > ");
                    pstrcat(string, buf_size, pFilterName);
                }
            }
            pFilter->Release();
        }

        pFilters->Release();
    }
}

void DShowDecoder::DeliverEndOfStream()
{
    pSource->DeliverEndOfStream();
}

#ifdef NOT_DONE
bool Ac3Decoder::LoadFiltersList(StringList & strList)
{
	IFilterMapper2 * pFilterMapper;

	HRESULT hr = CoCreateInstance(CLSID_FilterMapper2, NULL, CLSCTX_INPROC, IID_IFilterMapper2, (void **) &pFilterMapper);
	if (FAILED(hr))
	{
		return false;
	}

	GUID arrayInTypes[2];
	arrayInTypes[0] = MEDIATYPE_Audio;
	arrayInTypes[1] = MEDIASUBTYPE_DOLBY_AC3;
	GUID arrayOutTypes[2];
	arrayOutTypes[0] = MEDIATYPE_Audio;
	arrayOutTypes[1] = MEDIASUBTYPE_PCM;

	IEnumMoniker * pEnum = NULL;

	hr = pFilterMapper->EnumMatchingFilters(
		&pEnum,
		0,                  // Reserved.
		TRUE,               // Use exact match?
		MERIT_DO_NOT_USE+1, // Minimum merit.
		TRUE,               // At least one input pin?
		1,                  // Number of major type/subtype pairs for input.
		arrayInTypes,       // Array of major type/subtype pairs for input.
		NULL,               // Input medium.
		NULL,               // Input pin category.
		FALSE,              // Must be a renderer?
		TRUE,               // At least one output pin?
		1,                  // Number of major type/subtype pairs for output.
		arrayOutTypes,      // Array of major type/subtype pairs for output.
		NULL,               // Output medium.
		NULL);              // Output pin category.

	if (FAILED(hr))
	{
		pFilterMapper->Release();
		return false;
	}

	// Enumerate the monikers.
	IMoniker * pMoniker = NULL;
	ULONG cFetched;
	while(pEnum->Next(1, &pMoniker, &cFetched) == S_OK)
	{
		IPropertyBag * pPropBag = NULL;
		pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pPropBag);

		// To retrieve the friendly name of the filter, do the following:
		VARIANT varName;
		VariantInit(&varName);
		hr |= pPropBag->Read(L"FriendlyName", &varName, 0);
		if (SUCCEEDED(hr))
		{
			// Display the name in your UI somehow.
			std::wstring str( (wchar_t*)varName.bstrVal );
			strList.push_back( str );
		}
		VariantClear(&varName);

		// Clean up.
		pPropBag->Release();
		pMoniker->Release();
	}

	// Clean up
	pEnum->Release();
	pFilterMapper->Release();
	return (hr == S_OK);
}
#endif

#endif // CONFIG_WIN32
