/*
 * DTS DirectShow decoder (for win32 only)
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file DtsDecoder.cpp
 * DTS DShow entry point.
 */

#include <string>

#include <streams.h>

#include "ffconfig.h"
#include "common.h"
#include "avformat.h"

#if defined(CONFIG_WIN32)

#include "FilterUtil.h"
#include "DtsSourceCallBack.h"
#include "AudioSinkCallBack.h"
#include "DtsDecoder.h"

std::vector<whitelist> DtsDecoder::WhiteList;

SourceCallBack *DtsDecoder::GetSource(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	if (buf != NULL && !m_Frame.Eval(buf, buf_size))
		return NULL;

	// instantiate my source callback filter
	if ( m_Frame.IsValid() )
	{
		avctx->channels = m_Frame.Channels();
		avctx->sample_rate = m_Frame.SamplingRate();
	}
	
	return new DtsSourceCallBack( avctx->channels, avctx->sample_rate, avctx->extradata, avctx->extradata_size );
}

SinkCallBack *DtsDecoder::GetSink(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	if (buf != NULL && !m_Frame.Eval(buf, buf_size))
		return NULL;

	// instantiate my sink callback filter
	assert(avctx != NULL);
	if ( m_Frame.IsValid() )
	{
		avctx->channels = m_Frame.Channels();
		return new AudioSinkCallBack( *avctx, m_Frame.SamplingRate() );
	}
	else
		return new AudioSinkCallBack( *avctx, avctx->sample_rate );
}

#define _T(x) x

extern "C" {

#include "avcodec.h"

static int dts_decode_init(AVCodecContext *avctx)
{
	avctx->sample_fmt= SAMPLE_FMT_S16;
	avctx->priv_data = new DtsDecoder();

	return avctx->priv_data ? 0 : -1;
}

static int dts_decode_end(AVCodecContext *avctx)
{
	DtsDecoder *pDecoder = (DtsDecoder *)avctx->priv_data;
	delete pDecoder;
	avctx->priv_data = NULL;	
	return 0;
}

static int dts_decode_frame(AVCodecContext *avctx, void *outdata, int *outdata_size,
                uint8_t *buf, int buf_size)
{
	DtsDecoder *pDecoder = (DtsDecoder *)avctx->priv_data;

	if (!pDecoder->m_bGraphInit)
	{
        if (!pDecoder->InitGraph( avctx, buf, buf_size ))
			return -1;

		pDecoder->GraphInfo( avctx );

		if (pDecoder->m_Frame.IsValid())
		{
			avctx->bit_rate    = pDecoder->m_Frame.BitRate() * 1000;
			avctx->sample_rate = pDecoder->m_Frame.SamplingRate();
		}
	}

	*outdata_size = AVCODEC_MAX_AUDIO_FRAME_SIZE; // known size for all audio codec
	pDecoder->SetOutput( outdata, *outdata_size );

	if ( buf_size == 0 )
		pDecoder->DeliverEndOfStream();

	if (pDecoder->Deliver( buf, buf_size ))
		return buf_size;

	return -1;
}

static const char *dts_item_name(void*)
{
	return "dts_dshow";
}
static AVClass dts_dshow_class = {"dts_dshow", dts_item_name, NULL};

const char *dts_dshow_filter_string(void)
{
	static char *filter_string = NULL;
	if (filter_string == NULL)
	{		
		DtsDecoder *decoder = new DtsDecoder();
		AVCodecContext ctx;
		ctx.av_class = &dts_dshow_class;
		ctx.channels = 2;
		ctx.sample_rate = 44100;
		ctx.extradata = NULL;
		ctx.extradata_size = 0;

		if (!decoder->InitGraph( &ctx, NULL, 0 ))
			goto clean_exit;
		filter_string = (char*) av_malloc(256);
		decoder->GetFilterName( filter_string, 256 );
clean_exit:
		delete decoder;		
	}
	return filter_string;
}

int av_dts_dshow_whitelist(int argc, const whitelist *whitelists)
{
	int i;

	DtsDecoder::WhiteList.clear();

	for (i=0; i<argc; i++)
	{
		DtsDecoder::WhiteList.push_back(whitelists[i]);
	}

	return 0;
}

typedef DtsDecoder *DtsDecoderPrivate;

AVCodec dts_dshow_decoder = {
    "dts_dshow",
    CODEC_TYPE_AUDIO,
    CODEC_ID_DTS,
    sizeof(DtsDecoderPrivate),
    dts_decode_init,
    NULL,
    dts_decode_end,
    dts_decode_frame,
    /*.capabilities= */CODEC_CAP_DELAY,
};

};

#endif // CONFIG_WIN32
