/*
 * AC3 DirectShow decoder (for win32 only)
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file FilterUtil.cpp
 * DShow utilities.
 */

/***

  2001/03 Andrea Graziani

  Contains a collection of helper functions to load filters by name, 
  enumerate filters, find pins and connects pins.

***/

#define AV_DONT_OVERRIDE_STDC
#include <streams.h>

#include "FilterUtil.h"
#include "avcodec.h"

/**
 *
**/

// connect two filters using the named pins
//
HRESULT FilterUtil::Connect(IGraphBuilder * m_oGraph,
	IBaseFilter * pUpStreamFilter, IBaseFilter * pDownStreamFilter, 
	WCHAR * pszUpStreamPinName, WCHAR * pszDownStreamPinName)
{
	HRESULT hr;
	IPin *pOutPin=NULL, *pInPin=NULL;

	hr = FindPin(pUpStreamFilter, pszUpStreamPinName, &pOutPin, false);
	if (hr != S_OK) 
	{
		av_log(NULL, AV_LOG_ERROR, "FindPin(pUpStreamFilter, pszUpStreamPinName, &pOutPin) failed\n");
		goto clean_exit;
	}
	hr = FindPin(pDownStreamFilter, pszDownStreamPinName, &pInPin, true);
	if (hr != S_OK)
	{
		av_log(NULL, AV_LOG_ERROR, "FindPin(pDownStreamFilter, pszDownStreamPinName, &pInPin) failed\n");
		goto clean_exit;
	}

	hr = m_oGraph->Connect(pOutPin, pInPin);

	//create a New Segment on all pins
    IEnumFilters *pFilters;
    if (SUCCEEDED(m_oGraph->EnumFilters(&pFilters))) {
		IBaseFilter *pFilter;
		ULONG	n;

		while (pFilters->Next(1, &pFilter, &n) == S_OK) {
			IEnumPins *pEnumPins;
			IPin *pOutputPin;

			hr = pFilter->EnumPins( &pEnumPins );
			while( S_OK == pEnumPins->Next( 1, &pOutputPin, NULL ) )
			{
				pOutputPin->NewSegment(0, 0, 1.0);
				pOutputPin->Release();
			}
			pEnumPins->Release();
		}

		pFilters->Release();
	}

clean_exit:
	// only one reference kept in the connection needed
	if (pOutPin != NULL)
		pOutPin->Release();
	if (pInPin != NULL)
		pInPin->Release();

	return hr;
}

#if 0
// connect an output pin with a filter
//
HRESULT FilterUtil::Connect(IGraphBuilder * m_oGraph,
	IPin * pOutPin, IBaseFilter * pDownStreamFilter, WCHAR * pszDownStreamPinName)
{
	HRESULT hr;
	IPin * pInPin;

	FindPin(pDownStreamFilter, pszDownStreamPinName, &pInPin);

	hr = m_oGraph->Connect(pOutPin, pInPin);
	if (FAILED(hr)) {
		MessageBox(NULL, "I can't connect UpStream and DownStream Filters", "Error", MB_OK);
	}

	return hr;
}
#endif


// find a pin in a filter given the pin name
//
HRESULT FilterUtil::FindPin(IBaseFilter * pFilter, WCHAR * pszName, IPin ** ppPin, bool bInput)
{
	IEnumPins * pEnumPins = NULL;
	IPin * pPin = NULL;
	PIN_INFO PinInfo;
	ULONG nFetched;
	*ppPin = NULL;

	// get an enumerated list of the pins in this filter
	pFilter->EnumPins(&pEnumPins);

	while (pEnumPins->Next(1, &pPin, &nFetched) == S_OK) {
		pPin->QueryPinInfo(&PinInfo);
		// i have to release this since QueryPinInfo() adds a reference to the filter
		if (PinInfo.pFilter)
			PinInfo.pFilter->Release();

		if (pszName != NULL)
		{
			// return the IPin interface when find a match
			if (lstrcmpW(PinInfo.achName, pszName) == 0) {
				*ppPin = pPin;
				break;
			}
			LPWSTR PinString;
			pPin->QueryId(&PinString);
			if (lstrcmpW(PinString, pszName) == 0) {
				*ppPin = pPin;
				break;
			}
		}
		else
		{
			if (bInput && PinInfo.dir == PINDIR_INPUT) {
				*ppPin = pPin;
				break;
			}
			else if (!bInput && PinInfo.dir == PINDIR_OUTPUT) {
				*ppPin = pPin;
				break;
			}
		}

		pPin->Release();
	}

	pEnumPins->Release();
	return *ppPin != NULL ? S_OK : S_FALSE;
}


// check the HRESULT value
//
HRESULT FilterUtil::CheckHr(HRESULT hr, LPCSTR lpText)
{
	if (FAILED(hr)) {
		MessageBox(NULL, lpText, "Error", MB_OK);
	}

	return hr;
}


// add a filter in the filtergraph
//
void FilterUtil::AddFilter(IGraphBuilder * m_oGraph,
	IBaseFilter * pSourceFilter, LPCWSTR pName)
{
	HRESULT hr = m_oGraph->AddFilter(pSourceFilter, pName);
	CheckHr(hr, "Failed to add filter in the FilterGraph");

	// [Review] I'm not very sure that I have to release the filter
	pSourceFilter->Release();
}


// find a filter in the registry by name, and load it in the filtergraph
//
bool FilterUtil::LoadFilter(IGraphBuilder * m_oGraph,
	const GUID guid, const WCHAR * pszWFilterName, IBaseFilter ** pFilter)
{
#if 0
	wchar_t guidStr[50] = {'\0',};
	if (guid != GUID_NULL)
	{
		swprintf(guidStr, L"{%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X-}", guid.Data1, guid.Data2, guid.Data3, guid.Data4[0], guid.Data4[1]
			, guid.Data4[2], guid.Data4[3], guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);
	} 
	else
#endif
	if (pszWFilterName == NULL || pszWFilterName[0] == '\0')
		return false;

	BSTR pTmp;
	LONG lCount;

	bool bValidFilterFound = false;
	IFilterInfo * pFilterInfo = NULL;
	HRESULT hr;
#if 0  // doesn't work when adding the filter to the graph
	IFilterMapper2 *pMapper = NULL;
	IEnumMoniker *pEnum = NULL;

	CoCreateInstance(CLSID_FilterMapper2, NULL, CLSCTX_INPROC, IID_IFilterMapper2, (void **) &pMapper);
	    
	hr = pMapper->EnumMatchingFilters(
			&pEnum,
			0,                  // Reserved.
			FALSE,              // Use exact match?
			MERIT_DO_NOT_USE+1, // Minimum merit.
			TRUE,               // At least one input pin?
			0,                  // Number of major type/subtype pairs for input.
			NULL,               // Array of major type/subtype pairs for input.
			NULL,               // Input medium.
			NULL,               // Input pin category.
			FALSE,              // Must be a renderer?
			TRUE,               // At least one output pin?
			0,                  // Number of major type/subtype pairs for output.
			NULL,               // Array of major type/subtype pairs for output.
			NULL,               // Output medium.
			NULL);              // Output pin category.

	// Enumerate the monikers.
	IMoniker *pMoniker;
	ULONG cFetched;
	while(pEnum->Next(1, &pMoniker, &cFetched) == S_OK)
	{
		IPropertyBag *pPropBag;
		pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pPropBag);

		if (guid != GUID_NULL)
		{
			VARIANT varName, varID;
			VariantInit(&varID);
			hr = pPropBag->Read(L"CLSID", &varID, 0);
			if (SUCCEEDED(hr))
			{
				if (wcscmp((wchar_t*)varID.bstrVal, guidStr) == 0)
				{
					// Create an instance of the filter
					if (SUCCEEDED(pMoniker->BindToObject(NULL, NULL, IID_IBaseFilter, (void**)&pFilter)))
					{
						bValidFilterFound = true;

						VariantInit(&varName);
						hr = pPropBag->Read(L"FriendlyName", &varName, 0);
						if (SUCCEEDED(hr))
							m_oGraph->AddFilter(*pFilter, (wchar_t*)varName.bstrVal);
						else
							m_oGraph->AddFilter(*pFilter, pszWFilterName);
						// Clean up.
						VariantClear(&varName);
						VariantClear(&varID);
						pPropBag->Release();
						pMoniker->Release();
						break;
					}
				}
			}
			VariantClear(&varID);
		}
		else
		{
			VARIANT varName;
			VariantInit(&varName);
			hr = pPropBag->Read(L"FriendlyName", &varName, 0);
			if (SUCCEEDED(hr))
			{
				if (wcscmp((wchar_t*)varName.bstrVal, pszWFilterName) == 0)
				{
					// Create an instance of the filter
					pMoniker->BindToObject(NULL, NULL, IID_IBaseFilter, (void**)&pFilter);
					bValidFilterFound = true;

					m_oGraph->AddFilter(*pFilter, pszWFilterName);

					// Clean up.
					VariantClear(&varName);
					pPropBag->Release();
					pMoniker->Release();
					break;
				}
			}
			VariantClear(&varName);
		}

		// Clean up.
		pPropBag->Release();
		pMoniker->Release();
	}

	// Clean up
	pMapper->Release();
	pEnum->Release();
#endif

	if (guid != GUID_NULL)
	{
		if (SUCCEEDED(CoCreateInstance(guid, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void**) pFilter))) 
		{
			if (FAILED((*pFilter)->QueryInterface(IID_IFilterInfo, (void**)&pFilterInfo)))
//			if (FAILED((*pFilter)->QueryFilterInfo( &filter_info )))
				m_oGraph->AddFilter(*pFilter, pszWFilterName);
			else
			{
				pFilterInfo->get_Name(&pTmp);
				m_oGraph->AddFilter(*pFilter, pTmp);
				(*pFilter)->Release();
			}
			return true;
		}
	}

	IRegFilterInfo * pRegInfo = NULL;

	IMediaControl * pMC;
	IAMCollection * pMCollection;

	hr = m_oGraph->QueryInterface(IID_IMediaControl, (void **) &pMC);

	if (SUCCEEDED(hr)) {
		pMC->get_RegFilterCollection((IDispatch **) &pMCollection);
		pMC->Release();
	}

	// list of all registered directshow filters
	pMCollection->get_Count(&lCount);

	for (int i=0; i<lCount; i++)
	{
		pMCollection->Item(i, (IUnknown **) &pRegInfo);
		ASSERT(pRegInfo != NULL);
		pRegInfo->get_Name(&pTmp);

		// found the filter?
		if (lstrcmpW(pszWFilterName, pTmp) == 0)
		{
			hr = pRegInfo->Filter((IDispatch **) &pFilterInfo);
			if (pFilterInfo != NULL)
			{
				// get the filter and its properties
				hr = pFilterInfo->get_Filter((IUnknown **) pFilter);
				if (SUCCEEDED(hr)) {
					bValidFilterFound = true;
				}

				pFilterInfo->Release();
				break;
			}
		}

		pRegInfo->Release();
	}

	pMCollection->Release();

	return bValidFilterFound;
}

HRESULT AddToRot(IUnknown *pUnkGraph, DWORD & dwRegister )
{
#if defined(DEBUG)
    IMoniker * pMoniker;
    IRunningObjectTable *pROT;
    if (FAILED(GetRunningObjectTable(0, &pROT))) {
        return E_FAIL;
    }
    WCHAR wsz[256];
    wsprintfW(wsz, L"FilterGraph %08x pid %08x", (DWORD_PTR)pUnkGraph, GetCurrentProcessId());
    HRESULT hr = CreateItemMoniker(L"!", wsz, &pMoniker);
    if (SUCCEEDED(hr)) {
        hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pUnkGraph,
            pMoniker, &dwRegister);
        pMoniker->Release();
    }
    pROT->Release();
    return hr;
#else
	return S_OK;
#endif
}

void DelFromRot(DWORD dwRegister)
{
#if defined(DEBUG)
   IRunningObjectTable *pROT;
    if (SUCCEEDED(GetRunningObjectTable(0, &pROT))) {
        pROT->Revoke(dwRegister);
        pROT->Release();
    }
#endif
}