/*
 * RealVideo DirectShow decoder, using Gabest's filter (for win32 only)
 * Copyright (c) 2006 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file RealVideoDecoder.cpp
 * RealVideo DShow entry point.
 */

#include <string>

#include <streams.h>

#include "ffconfig.h"
#include "common.h"
#include "avformat.h"

#if defined(CONFIG_WIN32)

#include "FilterUtil.h"
#include "VideoSinkCallBack.h"
#include "RealVideoDecoder.h"

RealVideoDecoder::~RealVideoDecoder()
{
    if (frame_buffer != NULL)
        av_free(frame_buffer);
}

bool RealVideoDecoder::InitGraph( AVCodecContext *avctx, const uint8_t *buf, int buf_size )
{
	if (DShowDecoder::InitGraph(avctx, buf, buf_size))
	{
		assert(avctx != NULL);
		avctx->pix_fmt = pSink->avctx().pix_fmt;
		return true;
	}
	return false;
}

SourceCallBack *RealVideoDecoder::GetSource(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	return new RealVideoSourceCallback( avctx->width, avctx->height, avctx->codec_id, avctx->extradata, avctx->extradata_size );
}

SinkCallBack *RealVideoDecoder::GetSink(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	assert(m_avctx != NULL);
	return new VideoSinkCallBack( *m_avctx );
}

#define _T(x) x

extern "C" {

#include "avcodec.h"

static int rv_decode_init(AVCodecContext *avctx)
{
    avctx->priv_data = new RealVideoDecoder(avctx);

    return avctx->priv_data ? 0 : -1;
}

static int rv_decode_end(AVCodecContext *avctx)
{
    RealVideoDecoder *pDecoder = (RealVideoDecoder *)avctx->priv_data;
    delete pDecoder;
    avctx->priv_data = NULL;    
    return 0;
}

static int rv_decode_frame(AVCodecContext *avctx, void *outdata, int *outdata_size,
                uint8_t *buf, int buf_size)
{
    RealVideoDecoder *pDecoder = (RealVideoDecoder *)avctx->priv_data;

    if (!pDecoder->m_bGraphInit)
    {
        if (!pDecoder->InitGraph( avctx, buf, buf_size ))
            return -1;

        avctx->channels = 2;
        memset(&pDecoder->frame, 0, sizeof(pDecoder->frame));
        pDecoder->frame_len = avpicture_get_size(avctx->pix_fmt, avctx->width, avctx->height);
        pDecoder->frame_buffer = (uint8_t*) av_malloc(pDecoder->frame_len);

        pDecoder->GraphInfo( avctx );
    }

    pDecoder->SetOutput( pDecoder->frame_buffer, pDecoder->frame_len );

    if ( buf_size == 0 )
        pDecoder->DeliverEndOfStream();

    if (pDecoder->Deliver( buf, buf_size ))
    {
        avpicture_fill((AVPicture*)&pDecoder->frame, pDecoder->frame_buffer, avctx->pix_fmt,
                    avctx->width, avctx->height);
        *(AVFrame*)outdata = pDecoder->frame;
        *outdata_size = pDecoder->frame_len;
        return buf_size;
    }

    return -1;
}

static const char *rv_item_name(void*)
{
	return "rv_dshow";
}
static AVClass rv_dshow_class = {"rv_dshow", rv_item_name, NULL};

const char *rv_dshow_filter_string(void)
{
    static char *filter_string = NULL;
    if (filter_string == NULL)
    {        
        uint8_t rv_buf[5]= {0x13 ,0x10 ,0x56 ,0xe5 ,0x98};
        AVCodecContext ctx;
		ctx.av_class = &rv_dshow_class;
        ctx.channels = 2;
        ctx.sample_rate = 44100;
        ctx.extradata = rv_buf;
        ctx.extradata_size = 5;

        RealVideoDecoder *decoder = new RealVideoDecoder(&ctx);
        if (!decoder->InitGraph( &ctx, rv_buf, 5 ))
            goto clean_exit;
        filter_string = (char*) av_malloc(256);
        decoder->GetFilterName( filter_string, 256 );
clean_exit:
        delete decoder;        
    }
    return filter_string;
}

typedef RealVideoDecoder *RealVideoDecoderPrivate;

AVCodec rv40_dshow_decoder = {
    "rv40_dshow",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RV40,
    sizeof(RealVideoDecoderPrivate),
    rv_decode_init,
    NULL,
    rv_decode_end,
    rv_decode_frame,
    CODEC_CAP_DELAY,
};

AVCodec rv30_dshow_decoder = {
    "rv30_dshow",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RV30,
    sizeof(RealVideoDecoderPrivate),
    rv_decode_init,
    NULL,
    rv_decode_end,
    rv_decode_frame,
    CODEC_CAP_DELAY,
};

AVCodec rv20_dshow_decoder = {
    "rv20_dshow",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RV20,
    sizeof(RealVideoDecoderPrivate),
    rv_decode_init,
    NULL,
    rv_decode_end,
    rv_decode_frame,
    CODEC_CAP_DELAY,
};

AVCodec rv10_dshow_decoder = {
    "rv10_dshow",
    CODEC_TYPE_VIDEO,
    CODEC_ID_RV10,
    sizeof(RealVideoDecoderPrivate),
    rv_decode_init,
    NULL,
    rv_decode_end,
    rv_decode_frame,
    CODEC_CAP_DELAY,
};

};

#endif // CONFIG_WIN32
