/*
 * RealVideo DirectShow decoder, using Gabest's filter (for win32 only)
 * Copyright (c) 2006 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file SourceCallback.cpp
 * DShow RealVideo input pin.
 */

#define AV_DONT_OVERRIDE_STDC
#include <streams.h>

#include "avcodec.h"

#include "RealVideoSourceCallback.h"

/**
 *	source class
**/

// initialize a sourcecallbackstream object so there is a pin
// 
RealVideoSourceCallback::RealVideoSourceCallback( uint16_t width, uint16_t height, int codec, const void *codec_data, int codec_data_size )
: SourceCallBack( NAME("RealVideo Source Filter"), 0, 0 )
,m_CodecData(codec_data)
,m_CodecDataSize(codec_data_size)
,m_width(width)
,m_height(height)
,rvcodec(codec)
{
}

/**
 *	source output pin class
**/

RealVideoSourceCallbackPin::RealVideoSourceCallbackPin(RealVideoSourceCallback* pFilter, CCritSec* pLock, HRESULT* phr)
:	SourceCallBackPin( pFilter, pLock, phr)
{
}

HRESULT RealVideoSourceCallbackPin::CheckMediaType(const CMediaType * pmt) 
{
	if (*pmt->Type() != MEDIATYPE_Video)
		return S_FALSE;
	if (pmt->formattype != FORMAT_VideoInfo)
		return S_FALSE;

	RealVideoSourceCallback* pSrc = static_cast<RealVideoSourceCallback*>(m_pFilter);
	const VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER *)pmt->pbFormat;
	if ((pSrc->rvcodec == CODEC_ID_RV40 && pVih->bmiHeader.biCompression != 0x30345652) ||
		(pSrc->rvcodec == CODEC_ID_RV30 && pVih->bmiHeader.biCompression != 0x30335652) ||
		(pSrc->rvcodec == CODEC_ID_RV20 && pVih->bmiHeader.biCompression != 0x30325652) ||
		(pSrc->rvcodec == CODEC_ID_RV10 && pVih->bmiHeader.biCompression != 0x30315652))
		return S_FALSE;

	if (pVih->bmiHeader.biHeight != pSrc->m_height
		|| pVih->bmiHeader.biWidth != pSrc->m_width
//		|| pVih->AvgTimePerFrame != pSrc->m_FrameDuration
		)
		return S_FALSE;

	return S_OK;
} 

HRESULT RealVideoSourceCallbackPin::GetMediaType(int iPosition, CMediaType * pmt)
{
	if ( iPosition < 0 )
		return E_INVALIDARG;
	if ( iPosition != 0 )
		return VFW_S_NO_MORE_ITEMS;

	// the output pin force his mediatype to be used
	AM_MEDIA_TYPE * My = pmt;
	RealVideoSourceCallback *pSrc = static_cast<RealVideoSourceCallback*>(m_pFilter);
	VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER*)av_malloc(sizeof(VIDEOINFOHEADER)+pSrc->m_CodecDataSize);

	memset(pVih, 0, sizeof(*pVih));
//	pVih->AvgTimePerFrame = pSrc->m_FrameDuration;
	pVih->bmiHeader.biSize   = sizeof(pVih->bmiHeader);
	pVih->bmiHeader.biWidth = pSrc->m_width;
	pVih->bmiHeader.biHeight = pSrc->m_height;
    pVih->bmiHeader.biPlanes      = 3;
//	pVih->bmiHeader.biBitCount    = 1;
	if (pSrc->rvcodec == CODEC_ID_RV40)
		pVih->bmiHeader.biCompression = 0x30345652;
	else if (pSrc->rvcodec == CODEC_ID_RV30)
		pVih->bmiHeader.biCompression = 0x30335652;
	else if (pSrc->rvcodec == CODEC_ID_RV20)
		pVih->bmiHeader.biCompression = 0x30325652;
	else if (pSrc->rvcodec == CODEC_ID_RV10)
		pVih->bmiHeader.biCompression = 0x30315652;
    pVih->bmiHeader.biSizeImage   = pVih->bmiHeader.biWidth * pVih->bmiHeader.biHeight;

	memcpy(pVih+1, pSrc->m_CodecData, pSrc->m_CodecDataSize);

	memset(My, 0, sizeof(AM_MEDIA_TYPE));
	My->majortype = MEDIATYPE_Video;
	My->subtype = FOURCCMap(pVih->bmiHeader.biCompression);
	My->bFixedSizeSamples = FALSE;
	My->bTemporalCompression = FALSE;
	My->lSampleSize = pVih->bmiHeader.biSizeImage;
	My->formattype = FORMAT_VideoInfo;
	My->pUnk = NULL;
	My->cbFormat = sizeof(VIDEOINFOHEADER) + pSrc->m_CodecDataSize;
	My->pbFormat = (unsigned char *)CoTaskMemAlloc(My->cbFormat);
	memcpy(My->pbFormat, pVih, My->cbFormat);
	av_free(pVih);

	return S_OK;
}

HRESULT RealVideoSourceCallbackPin::DecideBufferSize(IMemAllocator *pAlloc, 
	ALLOCATOR_PROPERTIES *pProperties) 
{
	HRESULT hr;
	ALLOCATOR_PROPERTIES Actual;

	pProperties->cBuffers = 1;
	pProperties->cbBuffer = AVCODEC_MAX_AUDIO_FRAME_SIZE;
	pProperties->cbAlign = 1;
	pProperties->cbPrefix = 0;

	hr = pAlloc->SetProperties(pProperties, &Actual);
	if (FAILED(hr)) {
		return hr;
	}

	// verify that the allocator is able to allocate what requested
	if (Actual.cbBuffer < pProperties->cbBuffer) {
		return E_FAIL;
	}

	return S_OK;
}
