/*
 * DirectShow video capture for ffmpeg system
 * Copyright (c) 2005 DivX Inc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */

#define AV_DONT_OVERRIDE_STDC
#include <streams.h>

#include "FilterUtil.h"
#include "VideoSinkCallBack.h"

#include "VideoCapture.h"

DShowVideoCaptureGraph::DShowVideoCaptureGraph( AVFormatContext *avctx, const char *device )
	:DShowCaptureGraph(avctx,device)
	,m_device_input(NULL)
{
	char *separ = strrchr( m_devicename, '%');
	if (separ != NULL) {
		m_device_input = av_strdup( separ + 1);
		*separ = '\0';
	}
	else
		m_device_input = NULL;
}


DShowVideoCaptureGraph::~DShowVideoCaptureGraph()
{
	if ( m_device_input != NULL )
		av_free( m_device_input );
	DelFromRot( dwRotId );
}

bool DShowVideoCaptureGraph::CreateGraph( AVFormatParameters & avctx )
{
	// build the graph for the numbered device
    HRESULT hr;
    IBaseFilter *pSrcFilter=NULL;

    // Get DirectShow interfaces
    hr = GetInterfaces();
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Failed to get interfaces!  hr=0x%x\n"), hr);
		return false;
    }

    // Attach the filter graph to the capture graph
    hr = AttachFiltergraph();
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Failed to set capture filter graph!  hr=0x%x\n"), hr);
        return false;
    }

    // Use the system device enumerator and class enumerator to find
    // a video capture/preview device, such as a desktop USB video camera.
	hr = FindCaptureDevice(&pSrcFilter);
    if (FAILED(hr))
    {
        // Don't display a message because FindCaptureDevice will handle it
		return false;
    }
   
    // Add Capture filter to our graph.
	hr = g_pGraph->AddFilter(pSrcFilter, L"Video Capture");
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't add the capture filter to the graph!  hr=0x%x\r\n\r\n") 
            TEXT("If you have a working video capture device, please make sure\r\n")
            TEXT("that it is connected and is not being used by another application.\r\n\r\n")
            TEXT("The sample will now close."), hr);
        pSrcFilter->Release();
        return false;
    }

	// add a VideoSink
	assert(	m_Sink == NULL );
	m_Sink = new VideoSinkCallBack( m_avcctx );
	// query interface for the sink filter and add the
	// filter to the filtergraph
	IBaseFilter *pSinkFilter;
	if (FAILED(m_Sink->QueryInterface(IID_IBaseFilter, (void **) &pSinkFilter)))
		return false;
	if (FAILED(g_pGraph->AddFilter(pSinkFilter, L"Sink Filter")))
		return false;

    // Render the preview pin on the video capture filter and replace the Video Renderer with our sink
	g_pCapture->RenderStream(&PIN_CATEGORY_CAPTURE, &MEDIATYPE_Video, pSrcFilter, NULL, pSinkFilter);
    if (FAILED(hr))
    {
        av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't render the video capture stream.  hr=0x%x\r\n")
            TEXT("The capture device may already be in use by another application.\r\n\r\n")
            TEXT("The sample will now close."), hr);
        pSrcFilter->Release();
        return false;
    }

	if ( m_device_input != NULL && *m_device_input != '\0' )
	{
		IAMCrossbar *pVideoCrossbar = NULL;
		IBaseFilter *pFilter = NULL;
		IEnumFilters *pFilters;
		if (SUCCEEDED(g_pGraph->EnumFilters(&pFilters))) {
			ULONG	n;
			while (pFilters->Next(1, &pFilter, &n) == S_OK) {
				if (pVideoCrossbar == NULL)
				{
					pFilter->QueryInterface(IID_IAMCrossbar , (void **)&pVideoCrossbar);
					if (pVideoCrossbar != NULL)
						break;
				}
				pFilter->Release();
			}
			pFilters->Release();
		}

		if (pVideoCrossbar == NULL || pFilter == NULL)
		{
			av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't find the Crossbar to select pin %s\r\n"), m_device_input);
			pSrcFilter->Release();
			return false;
		}

		long outputs, inputs;
		if (FAILED(pVideoCrossbar->get_PinCounts(&outputs, &inputs)))
		{
			av_log(m_avctx, AV_LOG_ERROR, TEXT("Couldn't use the Crossbar to select pin %s\r\n"), m_device_input);
			pSrcFilter->Release();
			return false;
		}
		
		long pin, iOutputPin, iSelectPin;
		long related, phystype;
		for(pin=0; pin<outputs; ++pin) {
			if (FAILED(pVideoCrossbar->get_CrossbarPinInfo(FALSE, pin, &related, &phystype)))
				continue;
			
			if (phystype == PhysConn_Video_VideoDecoder) {
				iOutputPin = pin;
				break;
			}
		}

		PhysicalConnectorType InPhysType = GetPhysicalType(m_device_input);

		for(pin=0; pin<inputs; ++pin) {
			if (FAILED(pVideoCrossbar->get_CrossbarPinInfo(TRUE, pin, &related, &phystype)))
				continue;
			
			if (phystype == InPhysType) {
				iSelectPin = pin;
				break;
			}
		}

		pFilter->Release();
		pVideoCrossbar->Release();
	}

    // Now that the filter has been added to the graph and we have
    // rendered its stream, we can release this reference to the filter.
    pSrcFilter->Release();

	// find if there is a video renderer and remove it
	if (g_pGraph->FindFilterByName( L"Video Renderer", &pSrcFilter) == S_OK)
	{
		IPin * pInPin, *pConnectedPin;
		if (SUCCEEDED(FilterUtil::FindPin( pSrcFilter, L"Input", &pInPin, true )))
		{
			if (pInPin->ConnectedTo(&pConnectedPin) == S_OK)
			{
				pConnectedPin->Disconnect();
				pConnectedPin->Release();
			}
			pInPin->Disconnect();
			pInPin->Release();
		}
		g_pGraph->RemoveFilter( pSrcFilter );
		pSrcFilter->Release();
	}

	AddToRot( g_pGraph, dwRotId );
//while(1) Sleep(100);

	return true;
}

size_t DShowVideoCaptureGraph::GetDeviceList( StringList & out_list, AVFormatContext *avctx )
{
	return DShowCaptureGraph::GetDeviceList(out_list, CLSID_VideoInputDeviceCategory, avctx);
}

void DShowVideoCaptureGraph::GetDeviceCaps( const std::wstring & device, std::vector<VIDEOINFOHEADER> & out_list, AVFormatContext *avctx )
{
	out_list.clear();

	HRESULT hr;
	IBaseFilter *pSrc = FindDevice( device.c_str(), CLSID_VideoInputDeviceCategory, avctx );

	if (pSrc != NULL)
	{
		IEnumPins *pEnumPins;
		IPin *pOutputPin;

		hr = pSrc->EnumPins( &pEnumPins );
		if (FAILED(hr))
		{
			av_log(avctx, AV_LOG_ERROR, TEXT("GetDeviceCaps failed: no pin enumeration !\n"));
			pSrc->Release();
			goto done;
		}

		while( S_OK == pEnumPins->Next( 1, &pOutputPin, NULL ) )
		{
			PIN_INFO info;

			// TODO: add support for IID_IKsPropertySet to check for PIN_CATEGORY_CAPTURE 
			if( S_OK == pOutputPin->QueryPinInfo( &info ) )
			{
				if( info.pFilter ) info.pFilter->Release();
				if( info.dir == PINDIR_INPUT )
				{
					pOutputPin->Release();
					continue;
				}
				av_log(avctx, AV_LOG_DEBUG, "GetDeviceCaps: trying pin %S\n", info.achName );
			}

			IEnumMediaTypes *p_enum_mt;
			if ( SUCCEEDED(pOutputPin->EnumMediaTypes(&p_enum_mt)))
			{
				if ( SUCCEEDED(p_enum_mt->Reset()))
				{
					AM_MEDIA_TYPE *p_mt;
					ULONG          fetched;
					while( S_OK == p_enum_mt->Next( 1, &p_mt, &fetched ) )
					{
						if (pOutputPin->QueryAccept( p_mt ) == S_OK )
						{
							if( MEDIATYPE_Video == p_mt->majortype && FORMAT_VideoInfo == p_mt->formattype )
							{
								VIDEOINFOHEADER *pVinf = (VIDEOINFOHEADER *)p_mt->pbFormat;
								// verify the pixel format is one we can support
								if (VideoSinkCallBackPin::VideoFormatSubtype(p_mt->subtype) != PIX_FMT_NONE)
								{
									out_list.push_back(*pVinf);
								}
							}
						}
						DeleteMediaType( p_mt );
					}
				}
				p_enum_mt->Release();
			}
			pOutputPin->Release();
		}

		pEnumPins->Release();
		pSrc->Release();
	}

done:
	;
}

HRESULT DShowVideoCaptureGraph::FindCaptureDevice(IBaseFilter ** ppSrcFilter)
{
    HRESULT hr = S_OK;
	IBaseFilter * pSrc = NULL;

    if (!ppSrcFilter)
        return E_POINTER;
   
	wchar_t p_buf[256];
	int i_convert = MultiByteToWideChar(CP_ACP, 0, m_devicename, strlen(m_devicename)+1, p_buf, sizeof(p_buf));

	pSrc = FindDevice( p_buf, CLSID_VideoInputDeviceCategory, m_avctx );

	if (pSrc == NULL)
        return E_FAIL;

    // Copy the found filter pointer to the output parameter.
    // Do NOT Release() the reference, since it will still be used
    // by the calling function.
    *ppSrcFilter = pSrc;

    return hr;
}

void DShowVideoCaptureGraph::Configure( const wchar_t *device, HWND hwParent )
{
	// create a temporary graph to query the possible filters
	IBaseFilter *pSrc = FindDevice( device, CLSID_VideoInputDeviceCategory, NULL );
	if (pSrc == NULL)
		return;
	ISpecifyPropertyPages *pProp;
	HRESULT hr = pSrc->QueryInterface(IID_ISpecifyPropertyPages, (void **)&pProp);
	if (SUCCEEDED(hr)) 
	{
		// Get the filter's name and IUnknown pointer.
		FILTER_INFO FilterInfo;
		hr = pSrc->QueryFilterInfo(&FilterInfo); 
		IUnknown *pFilterUnk;
		pSrc->QueryInterface(IID_IUnknown, (void **)&pFilterUnk);

		// Show the page. 
		CAUUID caGUID;
		pProp->GetPages(&caGUID);
		pProp->Release();
		OleCreatePropertyFrame(
			hwParent,               // Parent window
			0, 0,                   // Reserved
			FilterInfo.achName,     // Caption for the dialog box
			1,                      // Number of objects (just the filter)
			&pFilterUnk,            // Array of object pointers. 
			caGUID.cElems,          // Number of property pages
			caGUID.pElems,          // Array of property page CLSIDs
			0,                      // Locale identifier
			0, NULL                 // Reserved
		);

		// Clean up.
		pFilterUnk->Release();
		CoTaskMemFree(caGUID.pElems);
	}
	pSrc->Release();
}

int DShowVideoCaptureGraph::ReadFrame(AVPacket *pkt)
{
	if ( g_psCurrent == Stopped )
	{
		if (FAILED(g_pMC->Run()))
			return -1;

		g_psCurrent = Running;
	}

	// prepare for 1 video frame
	int buffer_size = m_Stream->codec->width * m_Stream->codec->height * m_Stream->codec->bits_per_sample / 8;
    if (av_new_packet(pkt,buffer_size) < 0) {
        return AVERROR_NOMEM;
    }

	m_Sink->SetOutput( pkt->data, buffer_size );
	m_Sink->WaitOutput( );
	if  (buffer_size >= 0)
	{
		pkt->size = buffer_size;
		pkt->stream_index = m_Stream->index;
	}

	return buffer_size;
}

bool DShowVideoCaptureGraph::FillCodec( AVFormatParameters & avfctx )
{
	m_Stream = av_new_stream(m_avctx, 1);
	AVCodecContext & avctx = *m_Stream->codec;

	avctx.codec_type = CODEC_TYPE_VIDEO;
	avctx.codec_id = CODEC_ID_RAWVIDEO;
	avctx.pix_fmt = avfctx.pix_fmt;
	avctx.width = avfctx.width;
    avctx.height = avfctx.height;
	for (size_t i=0; i<sizeof(av_guid_pixfmt_table); i++)
	{
		if (avfctx.pix_fmt == av_guid_pixfmt_table[i].pix_fmt) {
			avctx.bits_per_sample = av_guid_pixfmt_table[i].bits_p_px;
			break;
		}
	}

	m_Stream->time_base = AV_TIME_BASE_Q;

	av_reduce(&avctx.time_base.num, &avctx.time_base.den, avfctx.time_base.num, avfctx.time_base.den, 60);
	double fps = double(avctx.time_base.den) / double(avctx.time_base.num);

	avctx.bit_rate = int(fps*avctx.width*avctx.height*avctx.bits_per_sample) >> 3;
	return true;
}

size_t DShowVideoCaptureGraph::GetFiltersWithProps( StringList & out_list, IAMTVTuner ** pTvTuner, IAMCrossbar ** pVideoCrossbar )
{
	out_list.clear();
	*pTvTuner = NULL;
	*pVideoCrossbar = NULL;

	IEnumFilters *pFilters;
    if (SUCCEEDED(g_pGraph->EnumFilters(&pFilters))) {
		IBaseFilter *pFilter;
		ULONG	n;
		while (pFilters->Next(1, &pFilter, &n) == S_OK) {
			ISpecifyPropertyPages *pProp;
			HRESULT hr = pFilter->QueryInterface(IID_ISpecifyPropertyPages, (void **)&pProp);
			if (SUCCEEDED(hr)) 
			{
				pProp->Release();

				// Get the filter's name and IUnknown pointer.
				FILTER_INFO FilterInfo;
				hr = pFilter->QueryFilterInfo(&FilterInfo);
				out_list.push_back( FilterInfo.achName );
				// The FILTER_INFO structure holds a pointer to the Filter Graph
				// Manager, with a reference count that must be released.
				if (FilterInfo.pGraph != NULL)
				{
					FilterInfo.pGraph->Release();
				}
			}
			if (*pTvTuner == NULL)
			{
				pFilter->QueryInterface(IID_IAMTVTuner, (void **)pTvTuner);
			}
			if (*pVideoCrossbar == NULL)
			{
				pFilter->QueryInterface(IID_IAMCrossbar , (void **)pVideoCrossbar);
			}
			pFilter->Release();
		}
		pFilters->Release();
	}

	return out_list.size();
}

bool DShowVideoCaptureGraph::FormatSupportsNTSC(ULONG format)
{
#define ALL_NTSC (AnalogVideo_NTSC_M|AnalogVideo_NTSC_M_J|AnalogVideo_NTSC_433)

	return ((format & ALL_NTSC) != 0);
}

bool DShowVideoCaptureGraph::FormatSupportsPAL(ULONG format)
{
#define ALL_PAL (AnalogVideo_PAL_B|AnalogVideo_PAL_D|AnalogVideo_PAL_H| \
	AnalogVideo_PAL_I|AnalogVideo_PAL_M|AnalogVideo_PAL_N|AnalogVideo_PAL_60)

	return ((format & ALL_PAL) != 0);
}

bool DShowVideoCaptureGraph::FormatSupportsSECAM(ULONG format)
{
#define ALL_SECAM (AnalogVideo_SECAM_B|AnalogVideo_SECAM_D|AnalogVideo_SECAM_G| \
	AnalogVideo_SECAM_H|AnalogVideo_SECAM_K|AnalogVideo_SECAM_K1|AnalogVideo_SECAM_L|AnalogVideo_SECAM_L1)

	return ((format & ALL_PAL) != 0);
}

extern "C" {

static int dshow_iread_video_header(struct AVFormatContext *cont, AVFormatParameters *ap)
{
	DShowVideoCaptureGraph **graph = (DShowVideoCaptureGraph **)cont->priv_data;
	
    if (!ap || ap->width <= 0 || ap->height <= 0 || ap->time_base.den <= 0)
        return -1;

	*graph = new DShowVideoCaptureGraph(cont, ap->device);

	(*graph)->m_avcctx.width = ap->width;
	(*graph)->m_avcctx.height = ap->height;
	(*graph)->m_avcctx.time_base = ap->time_base;

	if (!(*graph)->CreateGraph( *ap ))
	{
		delete *graph;
		*graph = NULL;
		return -1;
	}
	(*graph)->FillCodec( *ap );

	cont->pb.is_streamed = 1;

	return 0;
}

static int dshow_iread_packet(struct AVFormatContext *cont, AVPacket *pkt)
{
	DShowVideoCaptureGraph **graph = (DShowVideoCaptureGraph **)cont->priv_data;
	return (*graph)->ReadFrame( pkt );
}

static int dshow_iread_close(struct AVFormatContext *cont)
{
	DShowVideoCaptureGraph **graph = (DShowVideoCaptureGraph **)cont->priv_data;
	delete *graph;
	return 0;
}

static AVInputFormat dshow_vcap_format = {
    "dshowvcap",
    "DirectShow video capture",
    sizeof(DShowVideoCaptureGraph*),
    NULL, // read_probe
    dshow_iread_video_header,
    dshow_iread_packet,
    dshow_iread_close,
    NULL, // read_seek
    NULL, // read_timestamp <-
	AVFMT_NOFILE,
	/* .extensions =*/ "vcap",
};

int dshow_video_cap_init(void)
{
    av_register_input_format(&dshow_vcap_format);
    return 0;
}

}; // extern "C"
