/*
 * DirectShow Video Sink decoder (for win32 only)
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file VideoSinkCallBack.cpp
 * DShow YUV output pin.
 */

//#define DISABLE_PTS
#define AV_DONT_OVERRIDE_STDC
#include <streams.h>

#include "VideoSinkCallBack.h"

av_GUID_pix_fmt av_guid_pixfmt_table[8] =
{
	{ MEDIASUBTYPE_IYUV,   PIX_FMT_YUV420P, 12 },
	{ MEDIASUBTYPE_YUY2,   PIX_FMT_YUV422,  16 },
	{ MEDIASUBTYPE_UYVY,   PIX_FMT_UYVY422, 16 },
	{ MEDIASUBTYPE_RGB24,  PIX_FMT_RGB24,   24 },
	{ MEDIASUBTYPE_RGB32,  PIX_FMT_RGBA32,  32 },
	{ MEDIASUBTYPE_RGB555, PIX_FMT_RGB555,  16 },
	{ MEDIASUBTYPE_RGB565, PIX_FMT_RGB565,  16 },
	{ MEDIASUBTYPE_RGB8,   PIX_FMT_PAL8,    8  },
};


/**
 *	source class
**/

// initialize a sourcecallbackstream object so there is a pin
// 
VideoSinkCallBack::VideoSinkCallBack( AVCodecContext & avctx )
:SinkCallBack(avctx)
{
	m_Width = avctx.width;
	m_Height = avctx.height;
	m_FrameDuration = avctx.time_base.num;
	m_FrameDuration *= 10000000;
	m_FrameDuration /= avctx.time_base.den;
}

VideoSinkCallBack::~VideoSinkCallBack()
{
}

SinkCallBackPin *VideoSinkCallBack::CreateSinkPin(HRESULT & hr)
{
	return new VideoSinkCallBackPin(this, &m_Lock, &hr);
}


/**
 *	sink input pin class
**/

VideoSinkCallBackPin::VideoSinkCallBackPin(SinkCallBack* pFilter, CCritSec* pLock, HRESULT* phr)
:	SinkCallBackPin( pFilter, pLock, phr )
,m_FrameTime(0)
{
	m_hMemoryGiven = CreateEvent( NULL, FALSE, FALSE, NULL );
	m_hMemoryFilled = CreateEvent( NULL, FALSE, FALSE, NULL );
}

VideoSinkCallBackPin::~VideoSinkCallBackPin()
{
	CloseHandle( m_hMemoryGiven );
	CloseHandle( m_hMemoryFilled );
}

PixelFormat VideoSinkCallBackPin::VideoFormatSubtype(GUID & uid)
{
	PixelFormat result = PIX_FMT_NONE;
	FOURCCMap fourcc(&uid);
	
	for (size_t i=0; i<sizeof(av_guid_pixfmt_table); i++)
	{
		if (uid == av_guid_pixfmt_table[i].uid) {
			result = av_guid_pixfmt_table[i].pix_fmt;
			break;
		}
		if (uid == FOURCCMap(avcodec_pix_fmt_to_codec_tag(av_guid_pixfmt_table[i].pix_fmt))) {
			result = av_guid_pixfmt_table[i].pix_fmt;
			break;
		}
	}

	return result;
}

HRESULT VideoSinkCallBackPin::CheckMediaType(const CMediaType * pmt)
{
	if (*pmt->Type() != MEDIATYPE_Video)
		return S_FALSE;
	if (pmt->formattype != FORMAT_VideoInfo)
		return S_FALSE;

	VideoSinkCallBack* pSink = static_cast<VideoSinkCallBack*>(m_pFilter);
	const VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER *)pmt->pbFormat;
    FOURCCMap fourcc(pVih->bmiHeader.biCompression);
	if (pVih->bmiHeader.biHeight != pSink->m_Height
		|| pVih->bmiHeader.biWidth != pSink->m_Width
		|| (pVih->AvgTimePerFrame != 0 && pVih->AvgTimePerFrame != pSink->m_FrameDuration)
		|| VideoFormatSubtype(fourcc) == PIX_FMT_NONE)
		return S_FALSE;
    
	return S_OK;
} 

HRESULT VideoSinkCallBackPin::GetMediaType(int iPosition, CMediaType * pmt)
{
	if ( iPosition < 0 )
		return E_INVALIDARG;
	if ( iPosition > sizeof(av_guid_pixfmt_table) * 2 )
		return VFW_S_NO_MORE_ITEMS;

	// the output pin force his mediatypes to be used
	VideoSinkCallBack* pSink = static_cast<VideoSinkCallBack*>(m_pFilter);
	AM_MEDIA_TYPE * My = pmt;
	VIDEOINFOHEADER VIH;

	memset(&VIH, 0, sizeof(VIH));
	VIH.AvgTimePerFrame = pSink->m_FrameDuration;
	VIH.bmiHeader.biSize   = sizeof(VIH.bmiHeader);
	VIH.bmiHeader.biWidth = pSink->m_Width;
	VIH.bmiHeader.biHeight = pSink->m_Height;
    VIH.bmiHeader.biPlanes      = 3;
	VIH.bmiHeader.biBitCount    = av_guid_pixfmt_table[iPosition>>1].bits_p_px;
	VIH.bmiHeader.biCompression = avcodec_pix_fmt_to_codec_tag(av_guid_pixfmt_table[iPosition>>1].pix_fmt);
    VIH.bmiHeader.biSizeImage   = VIH.bmiHeader.biWidth * VIH.bmiHeader.biBitCount * VIH.bmiHeader.biHeight / 8;

	memset(My, 0, sizeof(AM_MEDIA_TYPE));
	My->majortype = MEDIATYPE_Video;
	if ( iPosition % 2 )
		My->subtype = FOURCCMap(avcodec_pix_fmt_to_codec_tag(av_guid_pixfmt_table[iPosition>>1].pix_fmt)); // PIX_FMT_YUV420P
	else
		My->subtype = av_guid_pixfmt_table[iPosition>>1].uid;
	My->bFixedSizeSamples = TRUE;
	My->bTemporalCompression = FALSE;
	My->lSampleSize = VIH.bmiHeader.biSizeImage;
	My->formattype = FORMAT_VideoInfo;
	My->pUnk = NULL;
	My->cbFormat = sizeof(VIDEOINFOHEADER);
	My->pbFormat = (unsigned char *)CoTaskMemAlloc(My->cbFormat);
	memcpy(My->pbFormat, &VIH, My->cbFormat);

	return S_OK;
}

HRESULT VideoSinkCallBackPin::SetMediaType(const CMediaType *pmt)
{
	assert(*pmt->Type() == MEDIATYPE_Video);
	assert(pmt->formattype == FORMAT_VideoInfo);
	VideoSinkCallBack* pSink = static_cast<VideoSinkCallBack*>(m_pFilter);
	const VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER *)pmt->pbFormat;
    FOURCCMap fourcc(pVih->bmiHeader.biCompression);

	pSink->avctx().pix_fmt = VideoFormatSubtype(fourcc);

	return S_OK;
}

void VideoSinkCallBackPin::SetOutput( void *outdata, int &outdata_size, int64_t *pts )
{
	m_pOutputBuf = outdata;
	m_piOutputSize = &outdata_size;
	m_pts = pts;

	SetEvent( m_hMemoryGiven );
}

void VideoSinkCallBackPin::WaitOutput()
{
	// wait for the buffer to be filled
	DWORD res = WaitForSingleObject( m_hMemoryFilled, 1000 );
	if ( res != WAIT_OBJECT_0 )
		*m_piOutputSize = -1;
}

HRESULT VideoSinkCallBackPin::Receive(IMediaSample * pSample)
{
	DWORD res = WaitForSingleObject( m_hMemoryGiven, 1500 );
	if ( res == WAIT_OBJECT_0 )
	{
		BYTE * pBuffer;
		LONG iSize = pSample->GetActualDataLength();

		if ( iSize > *m_piOutputSize )
		{
			// the sample is too big to be handled
			if (IsConnected())
				return S_OK;
			else
				return S_FALSE;
		}

#ifndef DISABLE_PTS
		if (m_pts != NULL)
		{
			REFERENCE_TIME refS,refE;
			if (pSample->GetTime(&refS, &refE) != VFW_E_SAMPLE_TIME_NOT_SET)
				*m_pts = refS;
			else
				*m_pts = AV_NOPTS_VALUE;
		}
#endif

		pSample->GetPointer(&pBuffer);
		memcpy( m_pOutputBuf , pBuffer, iSize );
		*m_piOutputSize = iSize;

		SetEvent( m_hMemoryFilled );

		return S_OK;
	}
	else
		return S_FALSE;
}
