/*
 * AC3 DirectShow decoder (for win32 only)
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file ac3_dshow.cpp
 * AC3 DShow entry point.
 */

#include <string>

#include <streams.h>

#include "ffconfig.h"
#include "common.h"
#include "avformat.h"

#if defined(CONFIG_WIN32)

#include "FilterUtil.h"
#include "Ac3SourceCallBack.h"
#include "AudioSinkCallBack.h"
#include "Ac3Decoder.h"

std::vector<whitelist> Ac3Decoder::WhiteList;

SourceCallBack *Ac3Decoder::GetSource(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	if (buf != NULL)
	{
		if ( buf_size < 7 )
			return NULL;

		m_Frame.Eval(buf);
	}

	// instantiate my source callback filter
	if ( m_Frame.IsValid() )
	{
		avctx->channels = m_Frame.Channels();
		avctx->sample_rate = m_Frame.SamplingRate();
	}

	return new Ac3SourceCallBack( avctx->channels, avctx->sample_rate );
}

SinkCallBack *Ac3Decoder::GetSink(AVCodecContext *avctx, const uint8_t *buf, int buf_size)
{
	if (buf != NULL)
	{
		if ( buf_size < 7 )
			return NULL;

		m_Frame.Eval(buf);
	}

	// instantiate my sink callback filter
	assert(avctx != NULL);
	if ( m_Frame.IsValid() )
	{
		avctx->channels = m_Frame.Channels();
		return new AudioSinkCallBack( *avctx, m_Frame.SamplingRate() );
	}
	else
		return new AudioSinkCallBack( *avctx, avctx->sample_rate );
}

#define _T(x) x

extern "C" {

#include "avcodec.h"

static int ac3_decode_init(AVCodecContext *avctx)
{
	avctx->sample_fmt= SAMPLE_FMT_S16;
	avctx->priv_data = new Ac3Decoder();

	return avctx->priv_data ? 0 : -1;
}

static int ac3_decode_end(AVCodecContext *avctx)
{
	Ac3Decoder *pDecoder = (Ac3Decoder *)avctx->priv_data;
	delete pDecoder;
	avctx->priv_data = NULL;	
	return 0;
}

static int ac3_decode_frame(AVCodecContext *avctx, void *outdata, int *outdata_size,
                uint8_t *buf, int buf_size)
{
	Ac3Decoder *pDecoder = (Ac3Decoder *)avctx->priv_data;

	if (!pDecoder->m_bGraphInit)
	{
        if (!pDecoder->InitGraph( avctx, buf, buf_size ))
			return -1;

		pDecoder->GraphInfo( avctx );

		avctx->bit_rate    = pDecoder->m_Frame.BitRate() * 1000;
		avctx->sample_rate = pDecoder->m_Frame.SamplingRate();
	}

	*outdata_size = AVCODEC_MAX_AUDIO_FRAME_SIZE; // known size for all audio codec
	pDecoder->SetOutput( outdata, *outdata_size );

	if ( buf_size == 0 )
		pDecoder->DeliverEndOfStream();

	if (pDecoder->Deliver( buf, buf_size ))
		return buf_size;

	return -1;
}

static const char *ac3_item_name(void*)
{
	return "ac3_dshow";
}
static AVClass ac3_dshow_class = {"ac3_dshow", ac3_item_name, NULL};

const char *ac3_dshow_filter_string(void)
{
	static char *filter_string = NULL;
	if (filter_string == NULL)
	{		
		Ac3Decoder *decoder = new Ac3Decoder();
		AVCodecContext ctx;
		ctx.av_class = &ac3_dshow_class;

		if (!decoder->InitGraph( &ctx, NULL, 0 ))
			goto clean_exit;
		filter_string = (char*) av_malloc(256);
		decoder->GetFilterName( filter_string, 256 );
clean_exit:
		delete decoder;		
	}
	return filter_string;
}

int av_ac3_dshow_whitelist(int argc, const whitelist *whitelists)
{
	int i;

	Ac3Decoder::WhiteList.clear();

	for (i=0; i<argc; i++)
	{
		Ac3Decoder::WhiteList.push_back(whitelists[i]);
	}

	return 0;
}

typedef Ac3Decoder *Ac3DecoderPrivate;

AVCodec ac3_dshow_decoder = {
    "ac3_dshow",
    CODEC_TYPE_AUDIO,
    CODEC_ID_AC3,
    sizeof(Ac3DecoderPrivate),
    ac3_decode_init,
    NULL,
    ac3_decode_end,
    ac3_decode_frame,
    /*.capabilities= */CODEC_CAP_DELAY,
};

};

#endif // CONFIG_WIN32
