/*
 * Generic AC3 frame parser
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */
/**
 * @file ac3frame.cpp
 */

#include <cassert>

#include "ac3frame.h"

const uint32_t AC3_Frame::SamplingRates[4] =
    //  00,   01,   10, 11
    {48000,44100,32000, 0};

const int16_t AC3_Frame::Bitrates[20] =
   { 32, 40, 48, 56, 64, 80, 96, 112, 128, 160, 192, 224, 256, 320, 384, 448, 512, 576, 640, -1};

AC3_Frame::AC3_Frame()
 :bValid(false)
 ,myChannels(AC3_LEFT_RIGHT)
 ,mySamplingRate(48000)
{}

/**
    \todo compute frame size
*/
bool AC3_Frame::Eval(const uint8_t Header[8])
{
    bValid = (Header[0] == 0x0B) && (Header[1] == 0x77);

    myBinaryValue[0] = Header[0];
    myBinaryValue[1] = Header[1];
    myBinaryValue[2] = Header[2];
    myBinaryValue[3] = Header[3];
    myBinaryValue[4] = Header[4];
    myBinaryValue[5] = Header[5];
    myBinaryValue[6] = Header[6];
    myBinaryValue[7] = Header[7];

//    big_uint16 bu16;
//    bu16.Eval(&Header[2]);
//    myCrc1 = bu16;
    myCrc1 = Header[2] << 8 | Header[3];
    
    // sampling rate
    mySamplingRate = SamplingRates[(Header[4] >> 6) & 0x03];

	if (mySamplingRate == 0)
		return false;

    myBitRate = Bitrates[(Header[4] >> 1) & 0x1F];

    myFrameSize = (myBitRate * 96000L) / mySamplingRate;
    if ((myBitRate * 96000L) % mySamplingRate)
    {
		if (Header[4] & 0x01)
			myFrameSize++;
    }
    myFrameSize <<= 1; // size in words

	myChannels = Ac3_Channel((Header[6] >> 5) & 0x07);
	myLfeChannel = (Header[7] & 0x20) == 1;

    return bValid;
}

/*
uint16_t AC3_Frame::SampleSize() const
{
    return 0; // unsupported format
}
*/

/**
    \todo simplify with more general cases
*/
uint16_t AC3_Frame::DataSize() const
{
    assert(bValid);

    return myFrameSize;
}

bool AC3_Frame::Compatible(const AC3_Frame & aFrame) const
{
    if (!aFrame.bValid)
	return false;

    if (mySamplingRate != aFrame.mySamplingRate)
	return false;

/*    if (myLayer != aFrame.myLayer)
	return false;

    if (Channels() != aFrame.Channels())
	return false;
*/
  return true;
}

uint16_t AC3_Frame::Channels() const
{
	uint16_t result = myLfeChannel ? 1: 0;
    switch (myChannels)
    {
	case AC3_1_AND_1:
		result += 2;
		break;
	case AC3_CENTER:
		result += 1;
		break;
	case AC3_LEFT_RIGHT:
		result += 2;
		break;
	case AC3_LEFT_CENTER_RIGHT:
		result += 3;
		break;
	case AC3_LEFT_RIGHT_SURROUND:
		result += 3;
		break;
	case AC3_LEFT_CENTER_RIGHT_SURROUND:
		result += 4;
		break;
	case AC3_LEFT_SURROUND_RIGHT_SURROUND:
		result += 4;
		break;
	case AC3_CENTER_LEFT_SURROUND_RIGHT_SURROUND:
		result += 5;
		break;
    }
    return result;
}
