/*
 * DirectShow capture io for ffmpeg system
 * Copyright (c) 2005 DivX Inc
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 */

#include <windows.h>
#include <dshow.h>
#include <streams.h>

#include <fcntl.h>

#include "SinkCallBack.h"
#include "FilterUtil.h"

#include "capture.h"

const DShowCaptureGraph::PhysicalCon DShowCaptureGraph::PhysConList[] =
{
	{ PhysConn_Video_Tuner,     "Video Tuner" }, 
    { PhysConn_Video_Composite, "Composite" }, 
    { PhysConn_Video_SVideo,    "S-Video" }, 
    { PhysConn_Video_RGB,       "RGB" }, 
    { PhysConn_Video_YRYBY,     "YRYBY" }, 
    { PhysConn_Video_SerialDigital,   "Serial Digital" }, 
    { PhysConn_Video_ParallelDigital, "Parallel Digital" }, 
    { PhysConn_Video_SCSI,            "SCSI Video" }, 
    { PhysConn_Video_AUX,       "Aux" }, 
    { PhysConn_Video_1394,      "Firewire Video" }, 
    { PhysConn_Video_USB,       "Video USB" }, 
    { PhysConn_Video_VideoDecoder, "Video Decoder" }, 
    { PhysConn_Video_VideoEncoder, "Video Encoder" }, 
    { PhysConn_Video_SCART,        "SCART plug" }, 

    { PhysConn_Audio_Tuner, "Audio Tuner" }, 
    { PhysConn_Audio_Line,  "Line-In" }, 
    { PhysConn_Audio_Mic,   "Microphone" }, 
    { PhysConn_Audio_AESDigital,   "AES Digital" }, 
    { PhysConn_Audio_SPDIFDigital, "S/PDIF" }, 
    { PhysConn_Audio_SCSI,  "SCSI Audio" }, 
    { PhysConn_Audio_AUX,   "Aux" }, 
    { PhysConn_Audio_1394,  "Firewire Audio" }, 
    { PhysConn_Audio_USB,   "Audio USB" }, 
    { PhysConn_Audio_AudioDecoder, "Audio Decoder" },
	{ PhysicalConnectorType(-1),   NULL            }
};

DShowCaptureGraph::DShowCaptureGraph( AVFormatContext *avctx, const char *device )
	:g_pMC(NULL)
	,g_pGraph(NULL)
	,g_pCapture(NULL)
	,g_psCurrent(Stopped)
	,m_avctx(avctx)
	,m_devicename(NULL)
	,m_OutputPinName(NULL)
	,m_Stream(NULL)
	,dwRotId(0)
	,m_Sink(NULL)
{
	m_devicename = av_strdup(device);
}


DShowCaptureGraph::~DShowCaptureGraph()
{
	CloseInterfaces();
	if (m_OutputPinName != NULL)
		CoTaskMemFree(m_OutputPinName);
	av_free(m_devicename);	
}

HRESULT DShowCaptureGraph::AttachFiltergraph()
{
	return g_pCapture->SetFiltergraph(g_pGraph);
}

HRESULT DShowCaptureGraph::GetInterfaces(void)
{
    HRESULT hr;

    // Create the filter graph
    hr = CoCreateInstance (CLSID_FilterGraph, NULL, CLSCTX_INPROC,
                           IID_IGraphBuilder, (void **) &g_pGraph);
    if (FAILED(hr))
        return hr;

    // Create the capture graph builder
    hr = CoCreateInstance (CLSID_CaptureGraphBuilder2 , NULL, CLSCTX_INPROC,
                           IID_ICaptureGraphBuilder2, (void **) &g_pCapture);
    if (FAILED(hr))
        return hr;
    
    // Obtain interfaces for media control
    hr = g_pGraph->QueryInterface(IID_IMediaControl,(LPVOID *) &g_pMC);
    if (FAILED(hr))
        return hr;

    return hr;
}

void DShowCaptureGraph::CloseInterfaces(void)
{
    // Stop capturing data
    if (g_psCurrent != Stopped && g_pMC)
        if (g_pMC->Stop() == S_OK)
		    g_psCurrent = Stopped;

    // Release DirectShow interfaces
	if (g_pMC != NULL)
		g_pMC->Release();
	if (g_pGraph != NULL)
		g_pGraph->Release();
	if (g_pCapture != NULL)
		g_pCapture->Release();
	delete m_Sink;
}

size_t DShowCaptureGraph::GetDeviceList( StringList & out_list, const GUID & DeviceType, AVFormatContext *avctx )
{
	out_list.clear();	

	HRESULT hr;
    IMoniker* pMoniker =NULL;
    ULONG cFetched;

    // Create the system device enumerator
    ICreateDevEnum* pDevEnum =NULL;

    hr = CoCreateInstance (CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC,
                           IID_ICreateDevEnum, (void **) &pDevEnum);
    if (FAILED(hr))
    {
        av_log(avctx, AV_LOG_ERROR, TEXT("Couldn't create system enumerator!  hr=0x%x\n"), hr);
        return 0;
    }

    // Create an enumerator for the video capture devices
    IEnumMoniker* pClassEnum = NULL;

	hr = pDevEnum->CreateClassEnumerator (DeviceType, &pClassEnum, 0);
	pDevEnum->Release();
    if (FAILED(hr))
    {
        av_log(avctx, AV_LOG_ERROR, TEXT("Couldn't create class enumerator!  hr=0x%x\n"), hr);
        return 0;
    }

    // If there are no enumerators for the requested type, then 
    // CreateClassEnumerator will succeed, but pClassEnum will be NULL.
    if (pClassEnum == NULL)
    {
		av_log(avctx, AV_LOG_ERROR, TEXT("No capture device was detected.\r\n\r\n")
                   TEXT("This sample requires a video capture device, such as a USB WebCam,\r\n")
                   TEXT("to be installed and working properly.  The sample will now close."),
                   TEXT("No Video Capture Hardware"), MB_OK | MB_ICONINFORMATION);
        return 0;
    }

    // Note that if the Next() call succeeds but there are no monikers,
    // it will return S_FALSE (which is not a failure).  Therefore, we
    // check that the return code is S_OK instead of using SUCCEEDED() macro.

    while( pClassEnum->Next( 1, &pMoniker, &cFetched ) == S_OK )
    {
        /* Getting the property page to get the device name */
        IPropertyBag *p_bag;
        hr = pMoniker->BindToStorage( 0, 0, IID_IPropertyBag, (void **)&p_bag );
		pMoniker->Release();
        if( SUCCEEDED(hr) )
        {
            VARIANT var;
            var.vt = VT_BSTR;
            hr = p_bag->Read( L"FriendlyName", &var, NULL );
			p_bag->Release();
			out_list.push_back((wchar_t*) var.bstrVal);
		}
	}
	pClassEnum->Release();

	return out_list.size();
}

IBaseFilter *DShowCaptureGraph::FindDevice( const wchar_t *device, const GUID & DeviceType, AVFormatContext *avctx )
{
	HRESULT hr;
    IBaseFilter *pResult = NULL;

    // Create the system device enumerator
    ICreateDevEnum* pDevEnum =NULL;
    IEnumMoniker* pClassEnum = NULL;
    IMoniker* pMoniker =NULL;
    ULONG cFetched;

    hr = CoCreateInstance (CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC,
                           IID_ICreateDevEnum, (void **) &pDevEnum);
    if (FAILED(hr))
    {
        av_log(avctx, AV_LOG_ERROR, TEXT("Couldn't create system enumerator!  hr=0x%x\n"), hr);
        goto done;
    }

    // Create an enumerator for the audio capture devices
	hr = pDevEnum->CreateClassEnumerator (DeviceType, &pClassEnum, 0);
	pDevEnum->Release();
    if (FAILED(hr))
    {
        av_log(avctx, AV_LOG_ERROR, TEXT("Couldn't create class enumerator!  hr=0x%x\n"), hr);
        goto done;
    }

    // If there are no enumerators for the requested type, then 
    // CreateClassEnumerator will succeed, but pClassEnum will be NULL.
    if (pClassEnum == NULL)
    {
		av_log(avctx, AV_LOG_ERROR, TEXT("No audio capture device was detected.\r\n\r\n")
                   TEXT("This sample requires an audio capture device, such as a USB WebCam,\r\n")
                   TEXT("to be installed and working properly.  The sample will now close."),
                   TEXT("No Audio Capture Hardware"), MB_OK | MB_ICONINFORMATION);
        goto done;
    }

    // Note that if the Next() call succeeds but there are no monikers,
    // it will return S_FALSE (which is not a failure).  Therefore, we
    // check that the return code is S_OK instead of using SUCCEEDED() macro.

    while( pClassEnum->Next( 1, &pMoniker, &cFetched ) == S_OK )
    {
        /* Getting the property page to get the device name */
        IPropertyBag *p_bag;
        hr = pMoniker->BindToStorage( 0, 0, IID_IPropertyBag, (void **)&p_bag );
        if( SUCCEEDED(hr) )
        {
            VARIANT var;
            var.vt = VT_BSTR;
            hr = p_bag->Read( L"FriendlyName", &var, NULL );
			p_bag->Release();
			if (wcscmp(device,var.bstrVal) == 0)
			{
				// Bind Moniker to a filter object
				hr = pMoniker->BindToObject(0,0,IID_IBaseFilter, (void**)&pResult);
				if (FAILED(hr))
				{
					av_log(avctx, AV_LOG_ERROR, TEXT("FindDevice failed: can't bind to filter !\n"));
				}
				pMoniker->Release();
				break;
			}
		}
		pMoniker->Release();
	}
	pClassEnum->Release();

done:	
	return pResult;
}

const char * DShowCaptureGraph::GetPhysicalInput(long physical)
{
	const char *result = NULL;
	const PhysicalCon *con = PhysConList;
	while ( con->physname != NULL )
	{
		if ( con->physcon == physical ) {
			result = con->physname;
			break;
		}
		con++;
	}
	return result;
}

PhysicalConnectorType DShowCaptureGraph::GetPhysicalType(const char *pin_type)
{
	PhysicalConnectorType result;
	const PhysicalCon *con = PhysConList;
	while ( con->physname != NULL )
	{
		if ( strcmp(pin_type, con->physname) == 0) {
			result = con->physcon;
			break;
		}
		con++;
	}
	return result;
}
