/*
 *  Copyright (C) 2005, DivX Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
*/
/**
    \file dtsframe.cpp
*/

#include <cassert>

#include "dtsframe.h"

const int32_t DTS_Frame::SamplingRates[16] =
    //  00,   01,   10, 11
    {-1, 8000, 16000, 32000, -1, -1, 11025, 22050, 44100, -1, -1, 12000, 24000, 48000, -1, -1};

const int16_t DTS_Frame::Bitrates[32] =
   { 32, 56, 64, 96, 112, 128, 192, 224, 256, 320, 384, 448, 512, 576, 640, 768, 960, 1024, 1152, 1280, 1344, 1408, 1411, 1472, 1536, 1920, 2048, 3072, 3840, -1, -1, -1 };

DTS_Frame::DTS_Frame()
 :bValid(false)
{}

bool DTS_Frame::Eval(const uint8_t *Header, size_t HeadSize)
{
	if (HeadSize < 10) {
		bValid = false;
		return false;
	}

    bValid = (Header[0] == 0x7F) && (Header[1] == 0xFE) && (Header[2] == 0x80) && (Header[3] == 0x01);

    myBinaryValue[0] = Header[0];
    myBinaryValue[1] = Header[1];
    myBinaryValue[2] = Header[2];
    myBinaryValue[3] = Header[3];
    myBinaryValue[4] = Header[4];//8
    myBinaryValue[5] = Header[5];//16
    myBinaryValue[6] = Header[6];//24
    myBinaryValue[7] = Header[7];//32
    myBinaryValue[8] = Header[8];//40
    myBinaryValue[9] = Header[9];//48

	myChannels = Dts_Channel((Header[7] & 0x0F) << 2 | (Header[8] >> 6));

    // sampling rate
    mySamplingRate = SamplingRates[(Header[8] & 0x3F) >> 2];

	if (mySamplingRate == -1)
		bValid = false;

    myBitRate = Bitrates[(Header[8] & 0x03) << 3 | Header[9] >> 5];

    return bValid;
}

/**
    \todo simplify with more general cases
*/
uint16_t DTS_Frame::DataSize() const
{
    assert(bValid);

    return myFrameSize;
}

bool DTS_Frame::Compatible(const DTS_Frame & aFrame) const
{
    if (!aFrame.bValid)
	return false;

    if (mySamplingRate != aFrame.mySamplingRate)
	return false;

  return true;
}

uint16_t DTS_Frame::Channels() const
{
	uint16_t result;
    switch (myChannels)
    {
	case DTS_1:
		result = 1;
		break;
	default:
	case DTS_1_AND_1:
	case DTS_LEFT_RIGHT:
	case DTS_LEFT_RIGHT_SUM_DIFF:
	case DTS_LEFT_RIGHT_TOTAL:
		result = 2;
		break;
	case DTS_CENTER_LEFT_RIGHT:
	case DTS_LEFT_RIGHT_SURROUND:
		result = 3;
		break;
	case DTS_CENTER_LEFT_RIGHT_SURROUND:
	case DTS_LEFT_RIGHT_SURROUND_LEFT_SURROUND_RIGHT:
		result = 4;
		break;
	case DTS_CENTER_LEFT_RIGHT_SURROUND_LEFT_SURROUND_RIGHT:
		result = 5;
		break;
	case DTS_CENTER_LEFT_CENTER_RIGHT_LEFT_RIGHT_SURROUND_LEFT_SURROUND_RIGHT:
	case DTS_CENTER_LEFT_RIGHT_LEFT_REAR_RIGHT_REAR_OVERHEAD:
	case DTS_CENTER_FRONT_CENTER_RIGHT_FRONT_LEFT_FRONT_RIGHT_LEFT_REAR_RIGHT_REAR:
		result = 6;
		break;
	case DTS_CENTER_LEFT_CENTER_CENTER_RIGHT_LEFT_RIGHT_SURROUND_LEFT_SURROUND_RIGHT:
		result = 7;
		break;
	case DTS_CENTER_LEFT_CENTER_RIGHT_LEFT_RIGHT_SURROUND_LEFT1_SURROUND_LEFT2_SURROUND_RIGHT1_SURROUND_RIGHT2:
	case DTS_CENTER_LEFT_CENTER_CENTER_RIGHT_LEFT_RIGHT_SURROUND_LEFT_SURROUND_SURROUND_RIGHT:
		result = 8;
		break;
    }
    return result;
}

uint16_t DTS_Frame::SamplingRate() const
{
	switch (mySamplingRate)
	{
	case 8000:
	case 16000:
	case 32000:
		return 32000;
	case 11025:
	case 22050:
	case 44100:
		return 44100;
	case 12000:
	case 24000:
	case 48000:
		return 48000;
	default:
		break;
	}
	return 0;
}
