/*
 * WMV decoder using the DMO DLLs (for win32 only)
 * Copyright (c) 2006 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02110-1301  USA
 *
 */

/**
 * @file wmvdmo.c
 * wmv codec.
 */

#include "avcodec.h"

#if defined(CONFIG_DMO_WMV) && defined(_WIN32)

#include <windows.h>
#include <dmo.h>
#include <dshow.h>

const static GUID wmv_GUID =          { 0x82d353df, 0x90bd, 0x4382, { 0x8b, 0xc2, 0x3f, 0x61, 0x92, 0xb7, 0x6e, 0x34 } };
const static GUID MEDIASUBTYPE_WMV1 = { 0x31564d57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WMV2 = { 0x32564D57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WMV3 = { 0x33564D57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WMVP = { 0x50564D57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WVP2 = { 0x32505657, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WMVR = { 0x52564D57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WMVA = { 0x41564D57, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };
const static GUID MEDIASUBTYPE_WVC1 = { 0x31435657, 0x0000, 0x0010, { 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71 } };

typedef interface i_mediabuffer
{
    struct IMediaBufferVtbl *lpVtbl;
} i_mediabuffer;

struct media_buffer {
    i_mediabuffer mb;
    int refcount;
    uint8_t *buffer;
    size_t buffer_size;
    size_t used_size;
} media_buffer;

static HRESULT __stdcall buffer_QueryInterface(i_mediabuffer * This,
            REFIID riid, void **ppvObject)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    if( !memcmp( riid, &IID_IUnknown, sizeof(GUID) ) || !memcmp( riid, &IID_IMediaBuffer, sizeof(GUID) ) )
    {
        bf->refcount++;
        *ppvObject = (void *)This;
        return NOERROR;
    }

    *ppvObject = NULL;
    return E_NOINTERFACE;
}

static ULONG __stdcall buffer_AddRef(i_mediabuffer * This)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    return bf->refcount++;
}

static ULONG __stdcall buffer_Release(i_mediabuffer * This)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    assert(bf->refcount > 0);
    if( --bf->refcount == 0 )
    {
        av_free( bf->mb.lpVtbl );
        av_free( bf );
    }

    return 0;
}

static HRESULT __stdcall buffer_SetLength(i_mediabuffer * This, DWORD cbLength)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    bf->used_size = cbLength;
    return S_OK;
}

static HRESULT __stdcall buffer_GetMaxLength(i_mediabuffer * This, DWORD *pcbMaxLength)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    if( !pcbMaxLength ) return E_POINTER;
    *pcbMaxLength = bf->buffer_size;
    return S_OK;
}

static HRESULT __stdcall buffer_GetBufferAndLength(i_mediabuffer * This,
    BYTE **ppBuffer,
    DWORD *pcbLength)
{
    struct media_buffer *bf=(struct media_buffer *)This;
    if (ppBuffer == NULL && pcbLength == NULL)
        return E_POINTER;
    if (ppBuffer) *ppBuffer = bf->buffer;
    if (pcbLength) *pcbLength = bf->used_size;
    return S_OK;
}

static struct media_buffer *create_buffer(uint8_t *buf, size_t buf_size)
{
    struct media_buffer *buffer = (struct media_buffer*)av_malloc(sizeof(media_buffer));
    if (!buffer)
        return NULL;

    buffer->mb.lpVtbl = (IMediaBufferVtbl *)av_malloc(sizeof(IMediaBufferVtbl));
    buffer->mb.lpVtbl->QueryInterface = buffer_QueryInterface;
    buffer->mb.lpVtbl->AddRef = buffer_AddRef;
    buffer->mb.lpVtbl->Release = buffer_Release;
    buffer->mb.lpVtbl->SetLength = buffer_SetLength;
    buffer->mb.lpVtbl->GetMaxLength = buffer_GetMaxLength;
    buffer->mb.lpVtbl->GetBufferAndLength = buffer_GetBufferAndLength;
    buffer->refcount = 1;
    buffer->buffer = buf;
    buffer->buffer_size = buf_size;
    buffer->used_size = 0;

    return buffer;
}

typedef struct WmvDMOContext {
    HMODULE wmv_dll;
    long (__stdcall *GetClass)(const GUID*, const GUID*, void** );
    IMediaObject *decoder;
    IClassFactory *classFactory;
    AVFrame frame;
} WmvDMOContext;

static int wmvdmo_decode_init(AVCodecContext *avctx)
{
    WmvDMOContext *s = avctx->priv_data;
    DMO_MEDIA_TYPE mediatype;
    VIDEOINFOHEADER *info;

    s->wmv_dll = LoadLibrary("wmvdecod.dll");

    if (s->wmv_dll == NULL)
    {
        av_log(avctx, AV_LOG_INFO, "wmvdecod.dll not found, trying wmvdmod.dll\n");
        s->wmv_dll = LoadLibrary("wmvdmod.dll");
    }

    if (s->wmv_dll == NULL)
    {
        av_log(avctx, AV_LOG_INFO, "wmvdmod.dll not found, trying wmv9dmod.dll\n");
        s->wmv_dll = LoadLibrary("wmv9dmod.dll");
    }

    if (s->wmv_dll == NULL)
    {
        av_log(avctx, AV_LOG_ERROR, "wmvdmod.dll or wmv9dmod.dll not found\n");
        return -1;
    }

    s->GetClass = (void*)GetProcAddress(s->wmv_dll, "DllGetClassObject");
    if (!s->GetClass)
    {
        av_log(avctx, AV_LOG_ERROR, "DllGetClassObject not found in the WMVDMO DLL\n");
        return -1;
    }

    if (s->GetClass(&wmv_GUID, &IID_IClassFactory, &s->classFactory ) != 0 || s->classFactory == NULL)
    {
        av_log(avctx, AV_LOG_ERROR, "This DLL can't decode WMV content\n");
        return -1;
    }

    if (s->classFactory->lpVtbl->CreateInstance(s->classFactory, 0, &IID_IMediaObject, &s->decoder) != 0 || s->decoder == NULL)
    {
        av_log(avctx, AV_LOG_ERROR, "Cannot create an instance of the WMV DMO decoder\n");
        return -1;
    }

    MoInitMediaType(&mediatype, sizeof(VIDEOINFOHEADER) + avctx->extradata_size);
    info = (void*) mediatype.pbFormat;
    memcpy( &info[1], avctx->extradata, avctx->extradata_size);

    mediatype.majortype = MEDIATYPE_Video;
    if (avctx->codec_id == CODEC_ID_WMV3)
        mediatype.subtype = MEDIASUBTYPE_WMV3;
    else if (avctx->codec_id == CODEC_ID_WMV2)
        mediatype.subtype = MEDIASUBTYPE_WMV2;
    else if (avctx->codec_id == CODEC_ID_VC1)
        mediatype.subtype = MEDIASUBTYPE_WVC1;
    else
        mediatype.subtype = MEDIASUBTYPE_WMV1;
    mediatype.formattype = FORMAT_VideoInfo;
    mediatype.bFixedSizeSamples = 0;
    mediatype.bTemporalCompression = 1;
    mediatype.lSampleSize = 0;

    info = (void*) mediatype.pbFormat;
    memset(info, 0, sizeof(*info));
    info->bmiHeader.biHeight = avctx->height;
    info->bmiHeader.biWidth = avctx->width;
    info->bmiHeader.biCompression = mediatype.subtype.Data1;
    info->bmiHeader.biPlanes = 1;
    info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER) + avctx->extradata_size;

    info->rcSource.left = info->rcSource.top = 0;
    info->rcSource.right = avctx->width;
    info->rcSource.bottom = avctx->height;
    info->rcTarget = info->rcSource;

    if (s->decoder->lpVtbl->SetInputType(s->decoder, 0, &mediatype, 0) != S_OK)
    {
        MoFreeMediaType(&mediatype);
        return -1;
    }

    MoFreeMediaType(&mediatype);
    MoInitMediaType(&mediatype, sizeof(VIDEOINFOHEADER));

    mediatype.majortype = MEDIATYPE_Video;
    mediatype.subtype = MEDIASUBTYPE_YUY2;
    mediatype.formattype = FORMAT_VideoInfo;
    mediatype.bTemporalCompression = 0;
    mediatype.bFixedSizeSamples = 1;

    info = (void*) mediatype.pbFormat;
    memset(info, 0, sizeof(*info));
    info->bmiHeader.biHeight = avctx->height;
    info->bmiHeader.biWidth = avctx->width;
    info->bmiHeader.biCompression = mediatype.subtype.Data1;
    info->bmiHeader.biBitCount = 16;
    info->bmiHeader.biPlanes = 1;
    info->bmiHeader.biSizeImage   = info->bmiHeader.biWidth * info->bmiHeader.biBitCount * info->bmiHeader.biHeight / 8;
    info->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);

    info->rcSource.left = info->rcSource.top = 0;
    info->rcSource.right = avctx->width;
    info->rcSource.bottom = avctx->height;
    info->rcTarget = info->rcSource;

    if (s->decoder->lpVtbl->SetOutputType(s->decoder, 0, &mediatype, 0) != S_OK)
    {
        MoFreeMediaType(&mediatype);
        return -1;
    }
    avctx->pix_fmt = PIX_FMT_YUV422;
    MoFreeMediaType(&mediatype);

    s->frame.data[0] = NULL;

    return 0;
}

static int wmvdmo_decode_end(AVCodecContext *avctx)
{
    WmvDMOContext *s = avctx->priv_data;

    if (s->frame.data[0])
        avctx->release_buffer(avctx, &s->frame);

    if (s->decoder)
        s->decoder->lpVtbl->Release(s->decoder);

    if (s->classFactory)
        s->classFactory->lpVtbl->Release(s->classFactory);

    if (s->wmv_dll != NULL)
        FreeLibrary(s->wmv_dll);
    return 0;
}

static int wmvdmo_decode_frame(AVCodecContext *avctx, void *outdata, int *outdata_size,
                uint8_t *buf, int buf_size)
{
    WmvDMOContext *s = avctx->priv_data;
    int len;
    HRESULT res;
    DWORD status;
    DMO_OUTPUT_DATA_BUFFER outp_buf;
    struct media_buffer *out_buffer, *in_buffer = create_buffer(buf, buf_size);

    assert(s->wmv_dll != NULL);

    in_buffer->used_size = buf_size;
    res = s->decoder->lpVtbl->ProcessInput(s->decoder, 0, (IMediaBuffer*)in_buffer, DMO_INPUT_DATA_BUFFERF_SYNCPOINT, 0, 0);

    buffer_Release((i_mediabuffer*)in_buffer);

    if (res != S_OK && res != S_FALSE)
    {
        av_log(avctx, AV_LOG_ERROR, "The DMO decoder doesn't accept the input sample\n");
        return -1;
    }

    if (res == S_OK)
    {
        // process output
        len = avpicture_get_size(avctx->pix_fmt, avctx->width, avctx->height);

        if (s->frame.data[0])
            avctx->release_buffer(avctx, &s->frame);

        if(avctx->get_buffer(avctx, &s->frame) < 0){
            av_log(avctx, AV_LOG_ERROR, "get_buffer() failed\n");
            return -1;
        }

        out_buffer = create_buffer(s->frame.data[0], len);
        memset(&outp_buf, 0, sizeof(outp_buf));
        outp_buf.pBuffer = (IMediaBuffer*)out_buffer;
        res = s->decoder->lpVtbl->ProcessOutput(s->decoder, DMO_PROCESS_OUTPUT_DISCARD_WHEN_NO_BUFFER, 1, &outp_buf, &status);
        avpicture_fill(&s->frame, out_buffer->buffer, avctx->pix_fmt,
                    avctx->width, avctx->height);
        *(AVFrame*)outdata = s->frame;
        s->frame.key_frame = 1; // because we don't know
        *outdata_size = out_buffer->used_size;
        buffer_Release((i_mediabuffer*)out_buffer);
    }

    return (res == S_OK)?buf_size:-1;
}

AVCodec wmv3_dmo_decoder = {
    "wmv3dmo",
    CODEC_TYPE_VIDEO,
    CODEC_ID_WMV3,
    sizeof(WmvDMOContext),
    wmvdmo_decode_init,
    NULL,
    wmvdmo_decode_end,
    wmvdmo_decode_frame,
    CODEC_CAP_DR1,
};

AVCodec wmv2_dmo_decoder = {
    "wmv2dmo",
    CODEC_TYPE_VIDEO,
    CODEC_ID_WMV2,
    sizeof(WmvDMOContext),
    wmvdmo_decode_init,
    NULL,
    wmvdmo_decode_end,
    wmvdmo_decode_frame,
    CODEC_CAP_DR1,
};

AVCodec wmv1_dmo_decoder = {
    "wmv1dmo",
    CODEC_TYPE_VIDEO,
    CODEC_ID_WMV1,
    sizeof(WmvDMOContext),
    wmvdmo_decode_init,
    NULL,
    wmvdmo_decode_end,
    wmvdmo_decode_frame,
    CODEC_CAP_DR1,
};

AVCodec wvc1_dmo_decoder = {
    "wvc1dmo",
    CODEC_TYPE_VIDEO,
    CODEC_ID_VC1,
    sizeof(WmvDMOContext),
    wmvdmo_decode_init,
    NULL,
    wmvdmo_decode_end,
    wmvdmo_decode_frame,
    CODEC_CAP_DR1,
};

#endif /* CONFIG_DMO_WMV */
