/*
 * DivX encoder (currently for win32 only)
 * Copyright (c) 2005 DivX, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include "ffconfig.h"
#include "common.h"

#ifdef CONFIG_DIVX

#ifdef CONFIG_WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <vfw.h>

#define DLL_NAME "divx.dll"
#define FUNC_NAME "DriverProc"

typedef LRESULT (WINAPI *DriverProc)(
        DWORD dwDriverId,
        HDRVR hDriver,
        UINT uMsg,
        LPARAM lParam1,
        LPARAM lParam2);

typedef struct divx_context {
    HMODULE dll;
    DWORD dwDriverId;
    DriverProc driverProc;
    char *state;
    int compressionStarted;
    int frameCount;
    BITMAPINFO *inputBmp;
    BITMAPINFO *outputBmp;    
} divx_context_t;

#endif

#include "avcodec.h"

#define MAX_CONFIG_SIZE 1024

static int divx_win32_encode_init(AVCodecContext *avctx)  {
    LRESULT result, size;
    divx_context_t *x;
    ICCOMPRESSFRAMES info;

    av_log(avctx, AV_LOG_DEBUG, "divx_encode_init()\n");

    if (avctx->pix_fmt != PIX_FMT_YUV420P) {
        av_log(avctx, AV_LOG_ERROR, "Color spaces other than 420p not supported\n");
        return -1;
    }

    x = avctx->priv_data;

    x->dll = LoadLibrary(DLL_NAME);
    
    if (x->dll == NULL) {
        av_log(avctx, AV_LOG_ERROR, "Could not load module '%s'\n",
            DLL_NAME);

        return -1;
    }
    
    av_log(avctx, AV_LOG_DEBUG, "Loaded '%s'\n", DLL_NAME);
    
    x->driverProc = (DriverProc) GetProcAddress(x->dll, FUNC_NAME);
    x->compressionStarted = 0;
    x->frameCount = 0;
    
    if (x->driverProc == NULL) {
        av_log(avctx, AV_LOG_ERROR, "Could not lookup function '%s'\n",
            FUNC_NAME);

        return -1;
    }
    
    av_log(avctx, AV_LOG_DEBUG, "Looked up '%s'\n", FUNC_NAME);
    
    x->dwDriverId = x->driverProc(0, 0, DRV_OPEN, 0, 0);

    if (x->dwDriverId == 0) {
        av_log(avctx, AV_LOG_ERROR, "Could not open driver\n");

        return -1;
    }        

    // Save state
    size = x->driverProc(x->dwDriverId, 0, ICM_GETSTATE, (LPARAM) NULL, 0);        

    if (size <= 0) {
        av_log(avctx, AV_LOG_ERROR, "Invalid state size: %d\n", (int) size);
        return -1;
    }
    
    x->state = (char *) av_malloc(size);
    
    result = x->driverProc(x->dwDriverId, 0, ICM_GETSTATE, (LPARAM) x->state, size);

    if (result != ICERR_OK) {
        av_log(avctx, AV_LOG_ERROR, "Could not obtain state\n");
        return -1;
    }
    
    av_log(avctx, AV_LOG_DEBUG, "State : %s\n", x->state);        

    if ( avctx->show_divx_config_ui) {
        x->driverProc(x->dwDriverId, 0, ICM_CONFIGURE, (LPARAM) GetDesktopWindow(), 0);
    } else if (avctx->opaque) {
        av_log(avctx, AV_LOG_INFO, "codec parameters: '%s'\n", avctx->opaque);
        x->driverProc(x->dwDriverId, 0, ICM_SETSTATE, (LPARAM) avctx->opaque, strlen(avctx->opaque) + 1);
    } else {
        char *newState = av_malloc(MAX_CONFIG_SIZE);
        char tmp[255];        
        
        memset(newState, 0, MAX_CONFIG_SIZE);

        if( avctx->flags & CODEC_FLAG_PASS1 ) {
            snprintf(tmp, 255, "-bvn1 %d ", avctx->bit_rate);
        } else if( avctx->flags & CODEC_FLAG_PASS2 ) {
            snprintf(tmp, 255, "-bvnn %d ", avctx->bit_rate);
        } else {
            snprintf(tmp, 255, "-bv1 %d ", avctx->bit_rate);
        }
        
        strncat(newState, tmp, MAX_CONFIG_SIZE);
        strncat(newState, "-vbv 4854000,3145728,2359296 -dir \"C:\\Program Files\\DivX\\DivX\" -b 1 -nf -profile=3", MAX_CONFIG_SIZE);

        av_log(avctx, AV_LOG_INFO, "codec custom parameters: '%s'\n", newState);
        x->driverProc(x->dwDriverId, 0, ICM_SETSTATE, (LPARAM) newState, strlen(newState) + 1);

        av_free(newState);
    }

    // Set frame rate    
    memset(&info, 0, sizeof(ICCOMPRESSFRAMES));

    info.dwRate = avctx->time_base.den;
    info.dwScale = avctx->time_base.num;

    result = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_FRAMES_INFO, (LPARAM) &info, sizeof(ICCOMPRESSFRAMES));

    if (result != ICERR_OK) {
        av_log(avctx, AV_LOG_ERROR, "Could not set frame rate\n");
        return -1;
    }

    x->inputBmp = av_malloc(sizeof(BITMAPINFO));
    memset(x->inputBmp, 0, sizeof(BITMAPINFO));

    x->inputBmp->bmiHeader.biSize            = sizeof(BITMAPINFOHEADER);
    x->inputBmp->bmiHeader.biWidth            = avctx->in_width;
    x->inputBmp->bmiHeader.biHeight            = avctx->in_height;
    x->inputBmp->bmiHeader.biPlanes            = 1;
    x->inputBmp->bmiHeader.biBitCount        = 0;
    x->inputBmp->bmiHeader.biCompression    = mmioFOURCC('I','4','2','0');
    x->inputBmp->bmiHeader.biSizeImage        = avctx->width * avctx->height * 3 / 2;

    size = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_GET_FORMAT, (LPARAM) x->inputBmp, (LPARAM) NULL);

    if (size <= 0) {
        av_log(avctx, AV_LOG_ERROR, "Invalid output format size: %i\n", (int) size);
        return -1;
    }

    x->outputBmp = av_malloc(size);
    memset(x->outputBmp, 0, size);

    result = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_GET_FORMAT, (LPARAM) x->inputBmp, (LPARAM) x->outputBmp);

    if (result != ICERR_OK) {
        av_log(avctx, AV_LOG_ERROR, "Could not obtain output format\n");
        return -1;
    }

    size = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_GET_SIZE, (LPARAM) x->inputBmp, (LPARAM) x->outputBmp);

    if (size <= 0) {
        av_log(avctx, AV_LOG_ERROR, "Could not obtain maximum output frame size\n");
        return -1;
    }    

    result = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_BEGIN, (LPARAM) x->inputBmp, (LPARAM) x->outputBmp);

    if (result != ICERR_OK) {
        av_log(avctx, AV_LOG_ERROR, "Could not start video compression\n");
        return -1;
    }
    
    x->compressionStarted = 1;        

    avctx->extradata = NULL;
    avctx->extradata_size = 0;
    avctx->codec_tag = ff_get_fourcc("DX50");
    avctx->coded_frame= avcodec_alloc_frame();
    
    return 0;
}

static int divx_win32_encode_frame(AVCodecContext *avctx,
                         unsigned char *frame, int buf_size, void *data) {    
    divx_context_t *x;
    LRESULT result;
    ICCOMPRESS iccompress;
    DWORD dwChunkId = 0;
    DWORD dwFlags = 0;
    AVPicture *picture;
    int size;
    char *picture_buf;
        
    av_log(avctx, AV_LOG_DEBUG,
            "divx_encode_frame(): %d, %d, %d\n", frame, buf_size, data);
    
    if (data == NULL)
        return -1;

    if (avctx->pix_fmt != PIX_FMT_YUV420P) {
        av_log(avctx, AV_LOG_ERROR, "Color spaces other than 420p not supported\n");
        return -1;
    }
        
    x = avctx->priv_data;    
    picture = data;

    size = (avctx->in_width * avctx->in_height * 3) / 2;
    picture_buf = av_malloc(size);
    
    avpicture_layout(picture, PIX_FMT_YUV420P, avctx->in_width, avctx->in_height,
                    picture_buf, size);

    iccompress.dwFlags = ICCOMPRESS_KEYFRAME; 
        iccompress.dwFlags = 0; 
    iccompress.lpbiOutput = &(x->outputBmp->bmiHeader); 
    iccompress.lpOutput = frame;
    iccompress.lpbiInput = &(x->inputBmp->bmiHeader); 
    iccompress.lpInput = picture_buf;
    iccompress.lpckid = &dwChunkId; 
    iccompress.lpdwFlags = &dwFlags; 
    iccompress.lFrameNum = x->frameCount; 
    iccompress.dwFrameSize = 0; 
    iccompress.dwQuality = 0; 
    iccompress.lpbiPrev = &(x->inputBmp->bmiHeader); 
    iccompress.lpPrev = NULL;
        
    result = x->driverProc(x->dwDriverId, 0, ICM_COMPRESS, (DWORD) &iccompress, sizeof(ICCOMPRESS));

    av_free(picture_buf);

    if (result != ICERR_OK) {
        av_log(avctx, AV_LOG_ERROR, "Could not compress video frame: %d\n", result);
        return -1;
    }    

    av_log(avctx, AV_LOG_DEBUG, "Compressed video frame size: %d\n", iccompress.lpbiOutput->biSizeImage);    
    
    // Ignore "lag for B-frame" byte
    if ((iccompress.lpbiOutput->biSizeImage == 1) && (frame[0] == 0x7f))
    {
        return 0;
    }    

    x->frameCount++;
    
    if (dwFlags & AVIIF_KEYFRAME) {
        avctx->coded_frame->key_frame = 1;
    } else {
        avctx->coded_frame->key_frame = 0;
    }        

    return iccompress.lpbiOutput->biSizeImage;    
}

static int divx_win32_encode_close(AVCodecContext *avctx) {
    divx_context_t *x = avctx->priv_data;        

    av_log(avctx, AV_LOG_DEBUG,
            "divx_encode_close()\n");        

    if (x->dwDriverId != 0) {
        if (x->compressionStarted) {
            x->driverProc(x->dwDriverId, 0, ICM_COMPRESS_END, 0, 0);
        }

        // Restore state
        x->driverProc(x->dwDriverId, 0, ICM_SETSTATE, (LPARAM) x->state, strlen(x->state) + 1);
        x->driverProc(x->dwDriverId, 0, DRV_CLOSE, 0, 0);
    }

    av_free(x->state);

    if (x->inputBmp != NULL)
        av_free(x->inputBmp);

    if (x->outputBmp != NULL)
        av_free(x->outputBmp);
    
    if (x->dll != NULL)
        FreeLibrary(x->dll);

    av_free(avctx->coded_frame);

    return 0;
}

static int divx_encode_init(AVCodecContext *avctx)  {
#ifdef CONFIG_WIN32
    return divx_win32_encode_init(avctx);
#else
#error "Support for Win32 only"
#endif
}

static int divx_encode_frame(AVCodecContext *avctx,
                         unsigned char *frame, int buf_size, void *data) {
#ifdef CONFIG_WIN32
    return divx_win32_encode_frame(avctx, frame, buf_size, data);
#else
#error "Support for Win32 only"
#endif
}

static int divx_encode_close(AVCodecContext *avctx) {
#ifdef CONFIG_WIN32
    return divx_win32_encode_close(avctx);
#else
#error "Support for Win32 only"
#endif
}

#ifdef CONFIG_DIVX_ENCODER

AVCodec divx_encoder = {
    "divx",
    CODEC_TYPE_VIDEO,
    CODEC_ID_DIVX,
    sizeof(divx_context_t),
    divx_encode_init,
    divx_encode_frame,
    divx_encode_close,
    NULL,
    /*.capabilities= */CODEC_CAP_DELAY,
#if LIBAVCODEC_VERSION_INT < ((50<<16)+(0<<8)+0)
    NULL,
#endif
    NULL,
    NULL,
    NULL,
    basic_pix_fmt,
    // .pix_fmts= (enum PixelFormat[]){PIX_FMT_YUV420P, -1}, 
};

#endif /* CONFIG_DIVX_ENCODER */

#endif /* CONFIG_DIVX */
