/*
 * Buffered file io for ffmpeg system
 * Copyright (c) 2001 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
#include "avformat.h"
#include <fcntl.h>
#ifndef CONFIG_WIN32
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#else
#include <io.h>
#define open(fname,oflag,pmode) open_utf8(fname,oflag,pmode)
#endif /* CONFIG_WIN32 */
#include <sys/stat.h>


#ifdef CONFIG_WIN32
int UpdateFromUTF8(const char *filename, wchar_t **widefilename)
{
    // find the size of the final UCS-2 string
    size_t i, j, _Length;
    wchar_t *newname;
    size_t namesize = strlen(filename);
    for (_Length=0, i=0; i<namesize; _Length++) {
        if ((filename[i] & 0x80) == 0) {
            i++;
        } else if ((filename[i] & 0x20) == 0) {
            i += 2;
        } else if ((filename[i] & 0x10) == 0) {
            i += 3;
        }
    }
    newname = *widefilename = av_malloc(i*sizeof(wchar_t));

    for (j=0, i=0; i<namesize; j++) {
        if ((filename[i] & 0x80) == 0) {
            newname[j] = filename[i];
            i++;
        } else if ((filename[i] & 0x20) == 0) {
            newname[j] = ((filename[i] & 0x1F) << 6) + (filename[i+1] & 0x3F);
            i += 2;
        } else if ((filename[i] & 0x10) == 0) {
            newname[j] = ((filename[i] & 0x0F) << 12) + ((filename[i+1] & 0x3F) << 6) + (filename[i+2] & 0x3F);
            i += 3;
        }
    }
    newname[j] = 0;
    return j;
}

int open_utf8(char *filename, int oflags, int pmode)
{
    wchar_t *p_buf = NULL;
    int filename_size = UpdateFromUTF8(filename, &p_buf);

    /* convert the UTF8 filename to WideCharacter */
    int result = _wopen(p_buf, oflags, pmode);
    av_free(p_buf);
    return result;
}
#endif

/* standard file protocol */

static int file_open(URLContext *h, const char *filename, int flags)
{
    int access;
    int fd;

    strstart(filename, "file:", &filename);

    if (flags & URL_RDWR) {
        access = O_CREAT | O_TRUNC | O_RDWR;
    } else if (flags & URL_WRONLY) {
        access = O_CREAT | O_TRUNC | O_WRONLY;
    } else {
        access = O_RDONLY;
    }
#if defined(CONFIG_WIN32) || defined(CONFIG_OS2) || defined(__CYGWIN__)
    access |= O_BINARY;
#endif
    fd = open(filename, access, 0666);
    if (fd < 0)
        return -ENOENT;
    h->priv_data = (void *)(size_t)fd;
    return 0;
}

static int file_read(URLContext *h, unsigned char *buf, int size)
{
    int fd = (size_t)h->priv_data;
    return read(fd, buf, size);
}

static int file_write(URLContext *h, unsigned char *buf, int size)
{
    int fd = (size_t)h->priv_data;
    return write(fd, buf, size);
}

/* XXX: use llseek */
static offset_t file_seek(URLContext *h, offset_t pos, int whence)
{
    int fd = (size_t)h->priv_data;
#if defined(CONFIG_WIN32) && !defined(__CYGWIN__)
    return _lseeki64(fd, pos, whence);
#else
    return lseek(fd, pos, whence);
#endif
}

static int file_close(URLContext *h)
{
    int fd = (size_t)h->priv_data;
    return close(fd);
}

URLProtocol file_protocol = {
    "file",
    file_open,
    file_read,
    file_write,
    file_seek,
    file_close,
};

/* pipe protocol */

static int pipe_open(URLContext *h, const char *filename, int flags)
{
    int fd;

    if (flags & URL_WRONLY) {
        fd = 1;
    } else {
        fd = 0;
    }
#if defined(CONFIG_WIN32) || defined(CONFIG_OS2) || defined(__CYGWIN__)
    setmode(fd, O_BINARY);
#endif
    h->priv_data = (void *)(size_t)fd;
    h->is_streamed = 1;
    return 0;
}

static int pipe_read(URLContext *h, unsigned char *buf, int size)
{
    int fd = (size_t)h->priv_data;
    return read(fd, buf, size);
}

static int pipe_write(URLContext *h, unsigned char *buf, int size)
{
    int fd = (size_t)h->priv_data;
    return write(fd, buf, size);
}

static int pipe_close(URLContext *h)
{
    return 0;
}

URLProtocol pipe_protocol = {
    "pipe",
    pipe_open,
    pipe_read,
    pipe_write,
    NULL,
    pipe_close,
};

/* multiple file protocol */

struct fileconcat_info {
    size_t filenum;
    size_t current;
    int *file_desc;
    int64_t *file_sizes;
};

static int fileconcat_open(URLContext *h, const char *fileuri, int flags)
{
    int access, filepos_idx;
    int fd;
    char *filename, *filepos;
    struct fileconcat_info *finfo;
#if defined(CONFIG_WIN32) && !defined(__CYGWIN__)
    struct __stat64 fstats;
#else
    struct stat fstats;
#endif

    h->priv_data = av_malloc(sizeof(struct fileconcat_info));
    finfo = (struct fileconcat_info *) h->priv_data;
    finfo->filenum = 0;
    finfo->file_desc = NULL;
    finfo->file_sizes = NULL;
    finfo->current = 0;

    strstart(fileuri, IO_MULTI_FILE_PROTO, &fileuri);

    filename = alloca(strlen(fileuri) + 1);
    strcpy(filename, fileuri);
    filepos = filename;

    while (strlen(filepos))
    {
        for (filepos_idx=0; filepos[filepos_idx] != IO_MULTI_FILE_SEPAR; filepos_idx++);
        filepos[filepos_idx] = '\0';

        if (flags & URL_RDWR) {
            access = O_CREAT | O_TRUNC | O_RDWR;
        } else if (flags & URL_WRONLY) {
            access = O_CREAT | O_TRUNC | O_WRONLY;
        } else {
            access = O_RDONLY;
        }
#if defined(CONFIG_WIN32) || defined(CONFIG_OS2) || defined(__CYGWIN__)
        access |= O_BINARY;
#endif
        fd = open(filepos, access, 0666);
        if (fd < 0)
            return -ENOENT;

        finfo->filenum++;
        finfo->file_desc = av_realloc(finfo->file_desc, sizeof(int) * finfo->filenum);
        finfo->file_desc[finfo->filenum-1] = fd;
#if defined(CONFIG_WIN32) && !defined(__CYGWIN__)
        _fstat64(fd, &fstats);
#else
        fstat(fd, &fstats);
#endif
        finfo->file_sizes = av_realloc(finfo->file_sizes, sizeof(int64_t) * finfo->filenum);
        finfo->file_sizes[finfo->filenum-1] = fstats.st_size;

        filepos = &filepos[filepos_idx+1];
    }

    return 0;
}

/* XXX: use llseek */

#if defined(CONFIG_WIN32) && !defined(__CYGWIN__)
# define portlseek(x,y,z)   _lseeki64(x,y,z)
#else
# define portlseek(x,y,z)   lseek(x,y,z)
#endif

static int fileconcat_read(URLContext *h, unsigned char *buf, int size)
{
    struct fileconcat_info *finfo = (struct fileconcat_info *) h->priv_data;
    int result, total=0;

    if (finfo->current >= finfo->filenum)
    {
        total = -1;
        size = 0;
    }

    while (size > 0)
    {
        assert(finfo->current < finfo->filenum);
        result = read(finfo->file_desc[finfo->current], buf, size);
        if (result < 0)
            return result;
        total += result;
        buf += result;
        size -= result;

        /* move to the next file to read the rest */
        if (size > 0)
        {
            finfo->current++;
            if (finfo->filenum <= finfo->current)
                break;
            portlseek(finfo->file_desc[finfo->current], 0, SEEK_SET);
        }
    }
    return total;
}

static offset_t fileconcat_seek(URLContext *h, offset_t pos, int whence)
{
    size_t i,j;
    offset_t result = -1;
    struct fileconcat_info *finfo = (struct fileconcat_info *) h->priv_data;
    switch (whence)
    {
        case SEEK_END:
            assert(pos == -1 || pos == 0); /* other cases not supported */
            result = portlseek(finfo->file_desc[finfo->filenum-1], -1, SEEK_END);
            if (result >= 0)
            {
                for (i=0; i<finfo->filenum-1; i++)
                    result += finfo->file_sizes[i];
            }
            finfo->current = finfo->filenum-1;
            break;
        case SEEK_CUR:
            /* get the final position */
            for (j=0; j<finfo->current && j<finfo->filenum; j++)
                pos += finfo->file_sizes[j];
            if (finfo->current < finfo->filenum)
                pos += portlseek(finfo->file_desc[finfo->current], 0, SEEK_CUR);
            /* fall through with the correct final position */
        case SEEK_SET:
            for (i=0; i<finfo->filenum; i++)
            {
                if (pos >= finfo->file_sizes[i])
                    pos -= finfo->file_sizes[i];
                else
                    break;
            }
            if (i >= finfo->filenum)
            {
                result = 0;
                for (j=0; j<finfo->filenum; j++)
                    result += finfo->file_sizes[j];
                finfo->current = finfo->filenum-1;
            }
            else
            {
                result = portlseek(finfo->file_desc[i], pos, SEEK_SET);
                if (result >= 0)
                {
                    for (j=0; j<i; j++)
                        result += finfo->file_sizes[j];
                }
                finfo->current = i;
            }
            break;
    }
    return result;
}

static int fileconcat_close(URLContext *h)
{
    size_t i;
    int res = 0;
    struct fileconcat_info *finfo = (struct fileconcat_info *) h->priv_data;
    if (finfo != NULL)
    {
        for(i=0; i<finfo->filenum; i++)
        {
            res |= close(finfo->file_desc[i]);
        }
        av_free(finfo->file_desc);
        av_free(finfo->file_sizes);
        av_free(finfo);
    }

    return res;
}

URLProtocol fileconcat_protocol = {
    "cat",
    fileconcat_open,
    fileconcat_read,
    NULL,
    fileconcat_seek,
    fileconcat_close,
};
